"""Type heartbeat.a, version 001"""

import json
from typing import Any
from typing import Dict
from typing import Literal

from gw.errors import GwTypeError
from pydantic import BaseModel
from pydantic import validator


class HeartbeatA001(BaseModel):
    MyHex: str  #
    YourLastHex: str  #
    TypeName: Literal["heartbeat.a"] = "heartbeat.a"
    Version: str = "001"

    def as_dict(self) -> Dict[str, Any]:
        d = self.dict()
        return d

    def as_type(self) -> str:
        return json.dumps(self.as_dict())


class HeartbeatA001_Maker:
    type_name = "heartbeat.a"
    version = "001"

    def __init__(self, my_hex: str, your_last_hex: str):
        self.tuple = HeartbeatA001(
            MyHex=my_hex,
            YourLastHex=your_last_hex,
            #
        )

    @classmethod
    def tuple_to_type(cls, tuple: HeartbeatA001) -> str:
        return tuple.as_type()

    @classmethod
    def type_to_tuple(cls, t: str) -> HeartbeatA001:
        try:
            d = json.loads(t)
        except TypeError:
            raise GwTypeError("Type must be string or bytes!")
        if not isinstance(d, dict):
            raise GwTypeError(f"Deserializing {t} must result in dict!")
        return cls.dict_to_tuple(d)

    @classmethod
    def dict_to_tuple(cls, d: dict[str, Any]) -> HeartbeatA001:
        d2 = dict(d)
        if "MyHex" not in d2.keys():
            raise GwTypeError(f"dict {d2} missing MyHex")
        if "YourLastHex" not in d2.keys():
            raise GwTypeError(f"dict {d2} missing YourLastHex")
        if "TypeName" not in d2.keys():
            raise GwTypeError(f"dict {d2} missing TypeName")

        return HeartbeatA001(
            MyHex=d2["MyHex"],
            YourLastHex=d2["YourLastHex"],
            TypeName=d2["TypeName"],
            Version="001",
        )



import logging
from typing import Dict
from typing import List
from typing import Optional

from gw.errors import DcError

from gwbase.data_classes.gps_point import GpsPoint
from gwbase.enums import GNodeRole
from gwbase.enums import GNodeStatus


LOG_FORMAT = (
    "%(levelname) -10s %(asctime)s %(name) -30s %(funcName) "
    "-35s %(lineno) -5d: %(message)s"
)
LOGGER = logging.getLogger(__name__)


class GNode:
    by_id: Dict[str, "GNode"] = {}
    by_alias: Dict[str, "GNode"] = {}

    def __new__(cls, g_node_id: str, *args, **kwargs) -> "GNode":  # type: ignore
        try:
            return cls.by_id[g_node_id]
        except KeyError:
            instance = super().__new__(cls)
            cls.by_id[g_node_id] = instance
            return instance

    def __init__(
        self,
        g_node_id: str,
        alias: str,
        status: GNodeStatus,
        role: GNodeRole,
        g_node_registry_addr: str,
        prev_alias: Optional[str] = None,
        gps_point_id: Optional[str] = None,
        ownership_deed_id: Optional[int] = None,
        ownership_deed_validator_addr: Optional[str] = None,
        owner_addr: Optional[str] = None,
        daemon_addr: Optional[str] = None,
        trading_rights_id: Optional[int] = None,
        scada_algo_addr: Optional[str] = None,
        scada_cert_id: Optional[int] = None,
        component_id: Optional[str] = None,
        display_name: Optional[str] = None,
    ):
        """GNodes are the basic building blocks of GridWorks

        Args:
            g_node_id (str): _description_
            alias (str): _description_
            status (GNodeStatus): _description_
            role (GNodeRole): _description_
            g_node_registry_addr (str): _description_
            prev_alias (Optional[str], optional): _description_. Defaults to None.
            gps_point_id (Optional[str], optional): _description_. Defaults to None.
            ownership_deed_id (Optional[int], optional): _description_. Defaults to None.
            ownership_deed_validator_addr (Optional[str], optional): _description_. Defaults to None.
            owner_addr (Optional[str], optional): _description_. Defaults to None.
            daemon_addr (Optional[str], optional): _description_. Defaults to None.
            trading_rights_id (Optional[int], optional): _description_. Defaults to None.
            scada_algo_addr ()(Optional[str], optional): _description_. Defaults to None.
            scada_cert_id (Optional[int], optional): _description_. Defaults to None.
            component_id (Optional[str], optional): _description_. Defaults to None.
            display_name (Optional[str], optional): _description_. Defaults to None.

        Raises:
            DcError: If status is not a GNodeStatus enum, or role is not a GNodeRole enum
        """
        self.g_node_id = g_node_id
        self.alias = alias
        if not isinstance(status, GNodeStatus):
            raise DcError(f"status {status} must be GNodeStatus, got {type(status)}")
        self.status: GNodeStatus = status
        if not isinstance(role, GNodeRole):
            raise DcError(f"role {role} must be  GNodeRole, got {type(role)}")
        self.role: GNodeRole = role
        self.g_node_registry_addr = g_node_registry_addr
        self.prev_alias = prev_alias
        self.gps_point_id = gps_point_id
        self.ownership_deed_id = ownership_deed_id
        self.ownership_deed_validator_addr = ownership_deed_validator_addr
        self.owner_addr = owner_addr
        self.daemon_addr = daemon_addr
        self.trading_rights_id = trading_rights_id
        self.scada_algo_addr = scada_algo_addr
        self.scada_cert_id = scada_cert_id
        self.component_id = component_id
        self.display_name = display_name
        self.__class__.by_alias[self.alias] = self

    def __repr__(self) -> str:
        rs = f"GNode Alias: {self.alias}, Role: {self.role.value}, Status: {self.status.value}"
        if self.ownership_deed_id and self.role == GNodeRole.TerminalAsset:
            rs += f", TaDeedIdx: {self.ownership_deed_id}"
        return rs

    def gps_point(self) -> Optional[GpsPoint]:
        if self.gps_point_id is None:
            return None
        return GpsPoint.by_id["gps_point_id"]

    def is_root(self) -> bool:
        alias_list = self.alias.split(".")
        alias_list.pop()
        if len(alias_list) == 0:
            return True
        return False

    @classmethod
    def active_g_nodes(cls) -> List["GNode"]:
        g_nodes = list(GNode.by_alias.values())
        return list(filter(lambda x: x.status == GNodeStatus.Active, g_nodes))

    @classmethod
    def parent_from_alias(cls, alias: str) -> Optional["GNode"]:
        """
        Returns:
            - GNode. If the parent as suggested by the alias exists as an
            Active BaseGNode, returns that.
            - None. If alias is one word long (i.e. root of world)
        """
        alias_list = alias.split(".")
        alias_list.pop()
        parent_alias = ".".join(alias_list)
        if parent_alias in list(map(lambda x: x.alias, cls.active_g_nodes())):
            return GNode.by_alias[parent_alias]
        return None

    def parent(self) -> Optional["GNode"]:
        """
        Raises: DcError if "natural" parent (as suggested by alias) is not Active,
        and either
            - prev_alias is None, OR
            - the parent as suggested by prev_alias is not Active and/or
            does not exist.

        Returns:
            BaseGNode.   Parent BaseGNode
              - If the parent as suggested by the alias exists as an
            Active GNode, returns that ("natural" parent)
              - Else, if the parent as suggested by the prev_alais exists
              as an active GNode, returns that.
            None.
              - If alias is one word long (i.e. root of world)



        """
        if self.is_root():
            return None
        natural_parent = self.parent_from_alias(self.alias)
        if natural_parent is not None:
            return natural_parent

        # alias may point to incorrect parent if getting updated
        if self.prev_alias is None:
            raise DcError(f"error finding parent for {self.alias}!")

        parent_pending_alias_update = self.parent_from_alias(self.prev_alias)
        if parent_pending_alias_update is None:
            raise DcError(f"error finding parent for {self.alias}!")
        return parent_pending_alias_update

    def children(self) -> List["GNode"]:
        """Returns the list of BaseGnodes identifying this node as parent."""
        return list(filter(lambda x: x.parent() == self, GNode.by_alias.values()))



"""Type heartbeat.a, version 100"""

import json
import logging
from typing import Any
from typing import Dict
from typing import Literal

from gw.errors import GwTypeError
from gw.utils import snake_to_camel
from pydantic import BaseModel
from pydantic import Field
from pydantic import validator


LOG_FORMAT = (
    "%(levelname) -10s %(asctime)s %(name) -30s %(funcName) "
    "-35s %(lineno) -5d: %(message)s"
)
LOGGER = logging.getLogger(__name__)


class HeartbeatA(BaseModel):
    """
    Used to check that an actor can both send and receive messages.

    Payload for direct messages sent back and forth between actors, for example a Supervisor
    and one of its subordinates. Update: MyHex and YourLatestHex must have format HexChar

    [More info](https://gridworks.readthedocs.io/en/latest/g-node-instance.html)
    """

    my_hex: str = Field(
        title="Hex character getting sent",
        default="0",
    )
    your_last_hex: str = Field(
        title="Last hex character received from heartbeat partner",
        default="0",
    )
    type_name: Literal["heartbeat.a"] = "heartbeat.a"
    version: Literal["100"] = "100"

    class Config:
        allow_population_by_field_name = True
        alias_generator = snake_to_camel

    @validator("my_hex")
    def _check_my_hex(cls, v: str) -> str:
        try:
            check_is_hex_char(v)
        except ValueError as e:
            raise ValueError(f"MyHex failed HexChar format validation: {e}")
        return v

    @validator("your_last_hex")
    def _check_your_last_hex(cls, v: str) -> str:
        try:
            check_is_hex_char(v)
        except ValueError as e:
            raise ValueError(f"YourLastHex failed HexChar format validation: {e}")
        return v

    def as_dict(self) -> Dict[str, Any]:
        """
        Translate the object into a dictionary representation that can be serialized into a
        heartbeat.a.100 object.

        This method prepares the object for serialization by the as_type method, creating a
        dictionary with key-value pairs that follow the requirements for an instance of the
        heartbeat.a.100 type. Unlike the standard python dict method,
        it makes the following substantive changes:
        - Enum Values: Translates between the values used locally by the actor to the symbol
        sent in messages.
        - Removes any key-value pairs where the value is None for a clearer message, especially
        in cases with many optional attributes.

        It also applies these changes recursively to sub-types.
        """
        d = {
            key: value
            for key, value in self.dict(
                by_alias=True, include=self.__fields_set__ | {"type_name", "version"}
            ).items()
            if value is not None
        }
        return d

    def as_type(self) -> bytes:
        """
        Serialize to the heartbeat.a.100 representation.

        Instances in the class are python-native representations of heartbeat.a.100
        objects, while the actual heartbeat.a.100 object is the serialized UTF-8 byte
        string designed for sending in a message.

        This method calls the as_dict() method, which differs from the native python dict()
        in the following key ways:
        - Enum Values: Translates between the values used locally by the actor to the symbol
        sent in messages.
        - - Removes any key-value pairs where the value is None for a clearer message, especially
        in cases with many optional attributes.

        It also applies these changes recursively to sub-types.

        Its near-inverse is HeartbeatA.type_to_tuple(). If the type (or any sub-types)
        includes an enum, then the type_to_tuple will map an unrecognized symbol to the
        default enum value. This is why these two methods are only 'near' inverses.
        """
        json_string = json.dumps(self.as_dict())
        return json_string.encode("utf-8")

    def __hash__(self):
        return hash((type(self),) + tuple(self.__dict__.values()))  # noqa


class HeartbeatA_Maker:
    type_name = "heartbeat.a"
    version = "100"

    @classmethod
    def tuple_to_type(cls, tuple: HeartbeatA) -> bytes:
        """
        Given a Python class object, returns the serialized JSON type object.
        """
        return tuple.as_type()

    @classmethod
    def type_to_tuple(cls, t: bytes) -> HeartbeatA:
        """
        Given a serialized JSON type object, returns the Python class object.
        """
        try:
            d = json.loads(t)
        except TypeError:
            raise GwTypeError("Type must be string or bytes!")
        if not isinstance(d, dict):
            raise GwTypeError(f"Deserializing <{t}> must result in dict!")
        return cls.dict_to_tuple(d)

    @classmethod
    def dict_to_tuple(cls, d: dict[str, Any]) -> HeartbeatA:
        """
        Deserialize a dictionary representation of a heartbeat.a.100 message object
        into a HeartbeatA python object for internal use.

        This is the near-inverse of the HeartbeatA.as_dict() method:
          - Enums: translates between the symbols sent in messages between actors and
        the values used by the actors internally once they've deserialized the messages.
          - Types: recursively validates and deserializes sub-types.

        Note that if a required attribute with a default value is missing in a dict, this method will
        raise a GwTypeError. This differs from the pydantic BaseModel practice of auto-completing
        missing attributes with default values when they exist.

        Args:
            d (dict): the dictionary resulting from json.loads(t) for a serialized JSON type object t.

        Raises:
           GwTypeError: if the dict cannot be turned into a HeartbeatA object.

        Returns:
            HeartbeatA
        """
        d2 = dict(d)
        if "MyHex" not in d2.keys():
            raise GwTypeError(f"dict missing MyHex: <{d2}>")
        if "YourLastHex" not in d2.keys():
            raise GwTypeError(f"dict missing YourLastHex: <{d2}>")
        if "TypeName" not in d2.keys():
            raise GwTypeError(f"TypeName missing from dict <{d2}>")
        if "Version" not in d2.keys():
            raise GwTypeError(f"Version missing from dict <{d2}>")
        if d2["Version"] != "100":
            LOGGER.debug(
                f"Attempting to interpret heartbeat.a version {d2['Version']} as version 100"
            )
            d2["Version"] = "100"
        return HeartbeatA(**d2)


def check_is_hex_char(v: str) -> None:
    """Checks HexChar format

    HexChar format: single-char string in '0123456789abcdefABCDEF'

    Args:
        v (str): the candidate

    Raises:
        ValueError: if v is not HexChar format
    """
    if not isinstance(v, str):
        raise ValueError(f"<{v}> must be a hex char, but not even a string")
    if len(v) > 1:
        raise ValueError(f"<{v}> must be a hex char, but not of len 1")
    if v not in "0123456789abcdefABCDEF":
        raise ValueError(f"<{v}> must be one of '0123456789abcdefABCDEF'")



"""Type g.node.gt, version 002"""

import json
import logging
from typing import Any
from typing import Dict
from typing import Literal
from typing import Optional

from gw.errors import GwTypeError
from gw.utils import snake_to_camel
from pydantic import BaseModel
from pydantic import Extra
from pydantic import Field
from pydantic import validator

from gwbase.data_classes.g_node import GNode
from gwbase.enums import GNodeRole
from gwbase.enums import GNodeStatus


LOG_FORMAT = (
    "%(levelname) -10s %(asctime)s %(name) -30s %(funcName) "
    "-35s %(lineno) -5d: %(message)s"
)
LOGGER = logging.getLogger(__name__)


class GNodeGt(BaseModel):
    """
    Used to send and receive updates about GNodes.

    GNodes are the building blocks of Gridworks. They have slowly-changing state that must be
    kept in sync across a distributed system. Therefore, they require a global registry to act
    as Single Source of Truth (SSoT). This class is used for that SSoT to share information
    with actors about their GNodes, and the GNodes that they will observe and communicate with.

    [More info](https://gridworks.readthedocs.io/en/latest/g-node.html)
    """

    g_node_id: str = Field(
        title="Immutable identifier for GNode",
    )
    alias: str = Field(
        title="Structured mutable identifier for GNode",
        description=(
            "The GNode Aliases are used for organizing how actors in Gridworks communicate. Together, "
            "they also encode the known topology of the electric grid."
            "[More info](https://gridworks.readthedocs.io/en/latest/g-node-alias.html)"
        ),
    )
    status: GNodeStatus = Field(
        title="Lifecycle indicator",
    )
    role: GNodeRole = Field(
        title="Role within Gridworks",
        description="[More info](https://gridworks.readthedocs.io/en/latest/g-node-role.html)",
    )
    g_node_registry_addr: str = Field(
        title="Algorand address for GNodeRegistry",
        description=(
            "For actors in a Gridworks world, the GNodeRegistry is the Single Source of Truth "
            "for existence and updates to GNodes."
            "[More info](https://gridworks.readthedocs.io/en/latest/g-node-registry.html)"
        ),
    )
    prev_alias: Optional[str] = Field(
        title="Previous GNodeAlias",
        description=(
            "As the topology of the grid updates, GNodeAliases will change to reflect that. This "
            "may happen a handful of times over the life of a GNode."
        ),
        default=None,
    )
    gps_point_id: Optional[str] = Field(
        title="Lat/lon of GNode",
        description=(
            "Some GNodes, in particular those acting as avatars for physical devices that are "
            "part of or are attached to the electric grid, have physical locations. These locations "
            "are used to help validate the grid topology."
        ),
        default=None,
    )
    ownership_deed_id: Optional[int] = Field(
        title="Algorand Id of ASA Deed",
        description=(
            "The Id of the TaDeed Algorand Standard Asset if the GNode is a TerminalAsset."
            "[More info](https://gridworks.readthedocs.io/en/latest/ta-deed.html)"
        ),
        default=None,
    )
    ownership_deed_validator_addr: Optional[str] = Field(
        title="Algorand address of Validator",
        description=(
            "Deeds are issued by the GNodeFactory, in partnership with third party Validators."
            "[More info](https://gridworks.readthedocs.io/en/latest/ta-validator.html)"
        ),
        default=None,
    )
    owner_addr: Optional[str] = Field(
        title="Algorand address of the deed owner",
        default=None,
    )
    daemon_addr: Optional[str] = Field(
        title="Algorand address of the daemon app",
        description=(
            "Some GNodes have Daemon applications associated to them to handle blockchain operations."
        ),
        default=None,
    )
    trading_rights_id: Optional[int] = Field(
        title="Algorand Id of ASA TradingRights",
        description="The Id of the TradingRights Algorand Standard Asset.",
        default=None,
    )
    scada_algo_addr: Optional[str] = Field(
        title="ScadaAlgoAddr",
        default=None,
    )
    scada_cert_id: Optional[int] = Field(
        title="ScadaCertId",
        default=None,
    )
    component_id: Optional[str] = Field(
        title="Unique identifier for GNode's Component",
        description=(
            "Used if a GNode is an avatar for a physical device. The serial number of a device "
            "is different from its make/model. The ComponentId captures the specific instance "
            "of the device."
        ),
        default=None,
    )
    display_name: Optional[str] = Field(
        title="Display Name",
        description="Sample: Keene Oak's Atomic TNode",
        default=None,
    )
    type_name: Literal["g.node.gt"] = "g.node.gt"
    version: Literal["002"] = "002"

    class Config:
        extra = Extra.allow
        allow_population_by_field_name = True
        alias_generator = snake_to_camel

    @validator("g_node_id")
    def _check_g_node_id(cls, v: str) -> str:
        try:
            check_is_uuid_canonical_textual(v)
        except ValueError as e:
            raise ValueError(
                f"GNodeId failed UuidCanonicalTextual format validation: {e}"
            )
        return v

    @validator("alias")
    def _check_alias(cls, v: str) -> str:
        try:
            check_is_left_right_dot(v)
        except ValueError as e:
            raise ValueError(f"Alias failed LeftRightDot format validation: {e}")
        return v

    @validator("g_node_registry_addr")
    def _check_g_node_registry_addr(cls, v: str) -> str:
        try:
            check_is_algo_address_string_format(v)
        except ValueError as e:
            raise ValueError(
                f"GNodeRegistryAddr failed AlgoAddressStringFormat format validation: {e}"
            )
        return v

    @validator("prev_alias")
    def _check_prev_alias(cls, v: Optional[str]) -> Optional[str]:
        if v is None:
            return v
        try:
            check_is_left_right_dot(v)
        except ValueError as e:
            raise ValueError(f"PrevAlias failed LeftRightDot format validation: {e}")
        return v

    @validator("gps_point_id")
    def _check_gps_point_id(cls, v: Optional[str]) -> Optional[str]:
        if v is None:
            return v
        try:
            check_is_uuid_canonical_textual(v)
        except ValueError as e:
            raise ValueError(
                f"GpsPointId failed UuidCanonicalTextual format validation: {e}"
            )
        return v

    @validator("ownership_deed_id")
    def _check_ownership_deed_id(cls, v: Optional[int]) -> Optional[int]:
        if v is None:
            return v
        try:
            check_is_positive_integer(v)
        except ValueError as e:
            raise ValueError(
                f"OwnershipDeedId failed PositiveInteger format validation: {e}"
            )
        return v

    @validator("ownership_deed_validator_addr")
    def _check_ownership_deed_validator_addr(cls, v: Optional[str]) -> Optional[str]:
        if v is None:
            return v
        try:
            check_is_algo_address_string_format(v)
        except ValueError as e:
            raise ValueError(
                f"OwnershipDeedValidatorAddr failed AlgoAddressStringFormat format validation: {e}"
            )
        return v

    @validator("owner_addr")
    def _check_owner_addr(cls, v: Optional[str]) -> Optional[str]:
        if v is None:
            return v
        try:
            check_is_algo_address_string_format(v)
        except ValueError as e:
            raise ValueError(
                f"OwnerAddr failed AlgoAddressStringFormat format validation: {e}"
            )
        return v

    @validator("daemon_addr")
    def _check_daemon_addr(cls, v: Optional[str]) -> Optional[str]:
        if v is None:
            return v
        try:
            check_is_algo_address_string_format(v)
        except ValueError as e:
            raise ValueError(
                f"DaemonAddr failed AlgoAddressStringFormat format validation: {e}"
            )
        return v

    @validator("trading_rights_id")
    def _check_trading_rights_id(cls, v: Optional[int]) -> Optional[int]:
        if v is None:
            return v
        try:
            check_is_positive_integer(v)
        except ValueError as e:
            raise ValueError(
                f"TradingRightsId failed PositiveInteger format validation: {e}"
            )
        return v

    @validator("scada_algo_addr")
    def _check_scada_algo_addr(cls, v: Optional[str]) -> Optional[str]:
        if v is None:
            return v
        try:
            check_is_algo_address_string_format(v)
        except ValueError as e:
            raise ValueError(
                f"ScadaAlgoAddr failed AlgoAddressStringFormat format validation: {e}"
            )
        return v

    @validator("scada_cert_id")
    def _check_scada_cert_id(cls, v: Optional[int]) -> Optional[int]:
        if v is None:
            return v
        try:
            check_is_positive_integer(v)
        except ValueError as e:
            raise ValueError(
                f"ScadaCertId failed PositiveInteger format validation: {e}"
            )
        return v

    @validator("component_id")
    def _check_component_id(cls, v: Optional[str]) -> Optional[str]:
        if v is None:
            return v
        try:
            check_is_uuid_canonical_textual(v)
        except ValueError as e:
            raise ValueError(
                f"ComponentId failed UuidCanonicalTextual format validation: {e}"
            )
        return v

    def as_dict(self) -> Dict[str, Any]:
        """
        Translate the object into a dictionary representation that can be serialized into a
        g.node.gt.002 object.

        This method prepares the object for serialization by the as_type method, creating a
        dictionary with key-value pairs that follow the requirements for an instance of the
        g.node.gt.002 type. Unlike the standard python dict method,
        it makes the following substantive changes:
        - Enum Values: Translates between the values used locally by the actor to the symbol
        sent in messages.
        - Removes any key-value pairs where the value is None for a clearer message, especially
        in cases with many optional attributes.

        It also applies these changes recursively to sub-types.
        """
        d = {
            key: value
            for key, value in self.dict(
                by_alias=True, include=self.__fields_set__ | {"type_name", "version"}
            ).items()
            if value is not None
        }
        del d["Status"]
        d["StatusGtEnumSymbol"] = GNodeStatus.value_to_symbol(self.status)
        del d["Role"]
        d["RoleGtEnumSymbol"] = GNodeRole.value_to_symbol(self.role)
        return d

    def as_type(self) -> bytes:
        """
        Serialize to the g.node.gt.002 representation.

        Instances in the class are python-native representations of g.node.gt.002
        objects, while the actual g.node.gt.002 object is the serialized UTF-8 byte
        string designed for sending in a message.

        This method calls the as_dict() method, which differs from the native python dict()
        in the following key ways:
        - Enum Values: Translates between the values used locally by the actor to the symbol
        sent in messages.
        - - Removes any key-value pairs where the value is None for a clearer message, especially
        in cases with many optional attributes.

        It also applies these changes recursively to sub-types.

        Its near-inverse is GNodeGt.type_to_tuple(). If the type (or any sub-types)
        includes an enum, then the type_to_tuple will map an unrecognized symbol to the
        default enum value. This is why these two methods are only 'near' inverses.
        """
        json_string = json.dumps(self.as_dict())
        return json_string.encode("utf-8")

    def __hash__(self):
        return hash((type(self),) + tuple(self.__dict__.values()))  # noqa


class GNodeGt_Maker:
    type_name = "g.node.gt"
    version = "002"

    @classmethod
    def tuple_to_type(cls, tuple: GNodeGt) -> bytes:
        """
        Given a Python class object, returns the serialized JSON type object.
        """
        return tuple.as_type()

    @classmethod
    def type_to_tuple(cls, t: bytes) -> GNodeGt:
        """
        Given a serialized JSON type object, returns the Python class object.
        """
        try:
            d = json.loads(t)
        except TypeError:
            raise GwTypeError("Type must be string or bytes!")
        if not isinstance(d, dict):
            raise GwTypeError(f"Deserializing <{t}> must result in dict!")
        return cls.dict_to_tuple(d)

    @classmethod
    def dict_to_tuple(cls, d: dict[str, Any]) -> GNodeGt:
        """
        Deserialize a dictionary representation of a g.node.gt.002 message object
        into a GNodeGt python object for internal use.

        This is the near-inverse of the GNodeGt.as_dict() method:
          - Enums: translates between the symbols sent in messages between actors and
        the values used by the actors internally once they've deserialized the messages.
          - Types: recursively validates and deserializes sub-types.

        Note that if a required attribute with a default value is missing in a dict, this method will
        raise a GwTypeError. This differs from the pydantic BaseModel practice of auto-completing
        missing attributes with default values when they exist.

        Args:
            d (dict): the dictionary resulting from json.loads(t) for a serialized JSON type object t.

        Raises:
           GwTypeError: if the dict cannot be turned into a GNodeGt object.

        Returns:
            GNodeGt
        """
        d2 = dict(d)
        if "GNodeId" not in d2.keys():
            raise GwTypeError(f"dict missing GNodeId: <{d2}>")
        if "Alias" not in d2.keys():
            raise GwTypeError(f"dict missing Alias: <{d2}>")
        if "StatusGtEnumSymbol" not in d2.keys():
            raise GwTypeError(f"StatusGtEnumSymbol missing from dict <{d2}>")
        value = GNodeStatus.symbol_to_value(d2["StatusGtEnumSymbol"])
        d2["Status"] = GNodeStatus(value)
        del d2["StatusGtEnumSymbol"]
        if "RoleGtEnumSymbol" not in d2.keys():
            raise GwTypeError(f"RoleGtEnumSymbol missing from dict <{d2}>")
        value = GNodeRole.symbol_to_value(d2["RoleGtEnumSymbol"])
        d2["Role"] = GNodeRole(value)
        del d2["RoleGtEnumSymbol"]
        if "GNodeRegistryAddr" not in d2.keys():
            raise GwTypeError(f"dict missing GNodeRegistryAddr: <{d2}>")
        if "TypeName" not in d2.keys():
            raise GwTypeError(f"TypeName missing from dict <{d2}>")
        if "Version" not in d2.keys():
            raise GwTypeError(f"Version missing from dict <{d2}>")
        if d2["Version"] != "002":
            LOGGER.debug(
                f"Attempting to interpret g.node.gt version {d2['Version']} as version 002"
            )
            d2["Version"] = "002"
        return GNodeGt(**d2)

    @classmethod
    def tuple_to_dc(cls, t: GNodeGt) -> GNode:
        if t.g_node_id in GNode.by_id.keys():
            dc = GNode.by_id[t.g_node_id]
        else:
            dc = GNode(
                g_node_id=t.g_node_id,
                alias=t.alias,
                status=t.status,
                role=t.role,
                g_node_registry_addr=t.g_node_registry_addr,
                prev_alias=t.prev_alias,
                gps_point_id=t.gps_point_id,
                ownership_deed_id=t.ownership_deed_id,
                ownership_deed_validator_addr=t.ownership_deed_validator_addr,
                owner_addr=t.owner_addr,
                daemon_addr=t.daemon_addr,
                trading_rights_id=t.trading_rights_id,
                scada_algo_addr=t.scada_algo_addr,
                scada_cert_id=t.scada_cert_id,
                component_id=t.component_id,
                display_name=t.display_name,
            )
        return dc

    @classmethod
    def dc_to_tuple(cls, dc: GNode) -> GNodeGt:
        return GNodeGt(
            g_node_id=dc.g_node_id,
            alias=dc.alias,
            status=dc.status,
            role=dc.role,
            g_node_registry_addr=dc.g_node_registry_addr,
            prev_alias=dc.prev_alias,
            gps_point_id=dc.gps_point_id,
            ownership_deed_id=dc.ownership_deed_id,
            ownership_deed_validator_addr=dc.ownership_deed_validator_addr,
            owner_addr=dc.owner_addr,
            daemon_addr=dc.daemon_addr,
            trading_rights_id=dc.trading_rights_id,
            scada_algo_addr=dc.scada_algo_addr,
            scada_cert_id=dc.scada_cert_id,
            component_id=dc.component_id,
            display_name=dc.display_name,
        )

    @classmethod
    def type_to_dc(cls, t: str) -> GNode:
        return cls.tuple_to_dc(cls.type_to_tuple(t))

    @classmethod
    def dc_to_type(cls, dc: GNode) -> str:
        return cls.dc_to_tuple(dc).as_type()

    @classmethod
    def dict_to_dc(cls, d: dict[Any, str]) -> GNode:
        return cls.tuple_to_dc(cls.dict_to_tuple(d))


def check_is_algo_address_string_format(v: str) -> None:
    """
    AlgoAddressStringFormat format: The public key of a private/public Ed25519
    key pair, transformed into an  Algorand address, by adding a 4-byte checksum
    to the end of the public key and then encoding in base32.

    Raises:
        ValueError: if not AlgoAddressStringFormat format
    """
    import algosdk

    at = algosdk.abi.AddressType()
    try:
        result = at.decode(at.encode(v))
    except Exception as e:
        raise ValueError(f"Not AlgoAddressStringFormat: {e}")


def check_is_left_right_dot(v: str) -> None:
    """Checks LeftRightDot Format

    LeftRightDot format: Lowercase alphanumeric words separated by periods, with
    the most significant word (on the left) starting with an alphabet character.

    Args:
        v (str): the candidate

    Raises:
        ValueError: if v is not LeftRightDot format
    """
    from typing import List

    try:
        x: List[str] = v.split(".")
    except:
        raise ValueError(f"Failed to seperate <{v}> into words with split'.'")
    first_word = x[0]
    first_char = first_word[0]
    if not first_char.isalpha():
        raise ValueError(
            f"Most significant word of <{v}> must start with alphabet char."
        )
    for word in x:
        if not word.isalnum():
            raise ValueError(f"words of <{v}> split by by '.' must be alphanumeric.")
    if not v.islower():
        raise ValueError(f"All characters of <{v}> must be lowercase.")


def check_is_positive_integer(v: int) -> None:
    """
    Must be positive when interpreted as an integer. Interpretation as an
    integer follows the pydantic rules for this - which will round down
    rational numbers. So 1.7 will be interpreted as 1 and is also fine,
    while 0.5 is interpreted as 0 and will raise an exception.

    Args:
        v (int): the candidate

    Raises:
        ValueError: if v < 1
    """
    v2 = int(v)
    if v2 < 1:
        raise ValueError(f"<{v}> is not PositiveInteger")


def check_is_uuid_canonical_textual(v: str) -> None:
    """Checks UuidCanonicalTextual format

    UuidCanonicalTextual format:  A string of hex words separated by hyphens
    of length 8-4-4-4-12.

    Args:
        v (str): the candidate

    Raises:
        ValueError: if v is not UuidCanonicalTextual format
    """
    try:
        x = v.split("-")
    except AttributeError as e:
        raise ValueError(f"Failed to split on -: {e}")
    if len(x) != 5:
        raise ValueError(f"<{v}> split by '-' did not have 5 words")
    for hex_word in x:
        try:
            int(hex_word, 16)
        except ValueError:
            raise ValueError(f"Words of <{v}> are not all hex")
    if len(x[0]) != 8:
        raise ValueError(f"<{v}> word lengths not 8-4-4-4-12")
    if len(x[1]) != 4:
        raise ValueError(f"<{v}> word lengths not 8-4-4-4-12")
    if len(x[2]) != 4:
        raise ValueError(f"<{v}> word lengths not 8-4-4-4-12")
    if len(x[3]) != 4:
        raise ValueError(f"<{v}> word lengths not 8-4-4-4-12")
    if len(x[4]) != 12:
        raise ValueError(f"<{v}> word lengths not 8-4-4-4-12")



import logging
from typing import Dict
from typing import List
from typing import Optional

from gw.errors import DcError

from gwbase.data_classes.g_node import GNode
from gwbase.data_classes.gps_point import GpsPoint
from gwbase.enums import GniStatus
from gwbase.enums import GNodeRole
from gwbase.enums import GNodeStatus
from gwbase.enums import StrategyName


LOG_FORMAT = (
    "%(levelname) -10s %(asctime)s %(name) -30s %(funcName) "
    "-35s %(lineno) -5d: %(message)s"
)
LOGGER = logging.getLogger(__name__)


class GNodeInstance:
    by_id: Dict[str, "GNodeInstance"] = {}
    by_alias: Dict[str, "GNodeInstance"] = {}

    def __new__(cls, g_node_instance_id: str, *args, **kwargs) -> "GNodeInstance":  # type: ignore
        try:
            return cls.by_id[g_node_instance_id]
        except KeyError:
            instance = super().__new__(cls)
            cls.by_id[g_node_instance_id] = instance
            return instance

    def __init__(
        self,
        g_node_instance_id: str,
        g_node: GNode,
        strategy: StrategyName,
        status: GniStatus,
        supervisor_container_id: str,
        start_time_unix_s: int,
        end_time_unix_s: int = 0,
        algo_address: Optional[str] = None,
    ):
        self.g_node_instance_id = g_node_instance_id
        if not isinstance(g_node, GNode):
            raise DcError(f"g_node {g_node} must be GNode, got {type(g_node)}")
        self.g_node: GNode = g_node
        if not isinstance(strategy, StrategyName):
            raise DcError(
                f"strategy {strategy} must be StrategyName, got {type(strategy)}"
            )

        self.strategy: StrategyName = strategy
        if not isinstance(status, GniStatus):
            raise DcError(f"status {status} must be GniStatus, got {type(status)}")
        self.status: GniStatus = status
        self.supervisor_container_id: str = supervisor_container_id
        self.start_time_unix_s: int = start_time_unix_s
        self.end_time_unix_s: int = end_time_unix_s
        self.algo_address: Optional[str] = algo_address

        self.__class__.by_alias[self.g_node.alias] = self

    #####################################################
    # With the exception of Status, all attributes of a GNode
    # come through as properties for the GNodeInstance. e.g.
    # gni.alias instead of gni.g_node.alias in the code
    #####################################################

    @property
    def g_node_id(self) -> str:
        return self.g_node.g_node_id

    @property
    def alias(self) -> str:
        return self.g_node.alias

    @property
    def role(self) -> GNodeRole:
        return self.g_node.role

    @property
    def g_node_registry_addr(self) -> str:
        return self.g_node.g_node_registry_addr

    @property
    def prev_alias(self) -> Optional[str]:
        return self.g_node.prev_alias

    @property
    def gps_point_id(self) -> Optional[str]:
        return self.g_node.gps_point_id

    @property
    def ownership_deed_id(self) -> Optional[int]:
        return self.g_node.ownership_deed_id

    @property
    def ownership_deed_validator_addr(self) -> Optional[str]:
        return self.g_node.ownership_deed_validator_addr

    @property
    def owner_addr(self) -> Optional[str]:
        return self.g_node.owner_addr

    @property
    def daemon_addr(self) -> Optional[str]:
        return self.g_node.daemon_addr

    @property
    def trading_rights_id(self) -> Optional[int]:
        return self.g_node.trading_rights_id

    @property
    def component_id(self) -> Optional[str]:
        return self.g_node.component_id

    @property
    def display_name(self) -> Optional[str]:
        return self.g_node.display_name

    def gps_point(self) -> Optional[GpsPoint]:
        if self.gps_point_id is None:
            return None
        return GpsPoint.by_id["gps_point_id"]

    def is_root(self) -> bool:
        alias_list = self.alias.split(".")
        alias_list.pop()
        if len(alias_list) == 0:
            return True
        return False

    @classmethod
    def active_gnis(cls) -> List["GNodeInstance"]:
        gnis = list(GNodeInstance.by_alias.values())
        return list(filter(lambda x: x.g_node.status == GNodeStatus.Active, gnis))

    @classmethod
    def parent_from_alias(cls, alias: str) -> Optional["GNodeInstance"]:
        """
        Returns:
            - GNode. If the parent as suggested by the alias exists as an
            Active BaseGNode, returns that.
            - None. If alias is one word long (i.e. root of world)
        """
        alias_list = alias.split(".")
        alias_list.pop()
        parent_alias = ".".join(alias_list)
        if parent_alias in list(map(lambda x: x.alias, cls.active_gnis())):
            return GNodeInstance.by_alias[parent_alias]
        return None

    def parent(self) -> Optional["GNodeInstance"]:
        """
        Raises: DcError if "natural" parent (as suggested by alias) is not Active,
        and either
            - prev_alias is None, OR
            - the parent as suggested by prev_alias is not Active and/or
            does not exist.
        Returns:
            BaseGNode.   Parent BaseGNode
              - If the parent as suggested by the alias exists as an
            Active GNode, returns that ("natural" parent)
              - Else, if the parent as suggested by the prev_alais exists
              as an active GNode, returns that.
            None.
              - If alias is one word long (i.e. root of world)
        """
        if self.is_root():
            return None
        natural_parent = self.parent_from_alias(self.alias)
        if natural_parent is not None:
            return natural_parent

        # alias may point to incorrect parent if getting updated
        if self.prev_alias is None:
            raise DcError(f"error finding parent for {self.alias}!")

        parent_pending_alias_update = self.parent_from_alias(self.prev_alias)
        if parent_pending_alias_update is None:
            raise DcError(f"error finding parent for {self.alias}!")
        return parent_pending_alias_update

    def children(self) -> List["GNodeInstance"]:
        """Returns the list of BaseGnodes identifying this node as parent"""
        return list(
            filter(lambda x: x.parent() == self, GNodeInstance.by_alias.values())
        )



"""Type super.starter, version 000"""

import json
import logging
from typing import Any
from typing import Dict
from typing import List
from typing import Literal

from gw.errors import GwTypeError
from gw.utils import snake_to_camel
from pydantic import BaseModel
from pydantic import Extra
from pydantic import Field
from pydantic import validator

from gwbase.types.g_node_instance_gt import GNodeInstanceGt
from gwbase.types.g_node_instance_gt import GNodeInstanceGt_Maker
from gwbase.types.supervisor_container_gt import SupervisorContainerGt
from gwbase.types.supervisor_container_gt import SupervisorContainerGt_Maker


LOG_FORMAT = (
    "%(levelname) -10s %(asctime)s %(name) -30s %(funcName) "
    "-35s %(lineno) -5d: %(message)s"
)
LOGGER = logging.getLogger(__name__)


class SuperStarter(BaseModel):
    """
    Used by world to seed a docker container with data needed to spawn and superviser GNodeInstances
    """

    supervisor_container: SupervisorContainerGt = Field(
        title="Key data about the docker container",
    )
    gni_list: List[GNodeInstanceGt] = Field(
        title="List of GNodeInstances (Gnis) run in the container",
    )
    alias_with_key_list: List[str] = Field(
        title="Aliases of Gnis that own Algorand secret keys",
    )
    key_list: List[str] = Field(
        title="Algorand secret keys owned by Gnis",
    )
    type_name: Literal["super.starter"] = "super.starter"
    version: Literal["000"] = "000"

    class Config:
        extra = Extra.allow
        allow_population_by_field_name = True
        alias_generator = snake_to_camel

    @validator("alias_with_key_list")
    def _check_alias_with_key_list(cls, v: List[str]) -> List[str]:
        for elt in v:
            try:
                check_is_left_right_dot(elt)
            except ValueError as e:
                raise ValueError(
                    f"AliasWithKeyList element {elt} failed LeftRightDot format validation: {e}"
                )
        return v

    def as_dict(self) -> Dict[str, Any]:
        """
        Translate the object into a dictionary representation that can be serialized into a
        super.starter.000 object.

        This method prepares the object for serialization by the as_type method, creating a
        dictionary with key-value pairs that follow the requirements for an instance of the
        super.starter.000 type. Unlike the standard python dict method,
        it makes the following substantive changes:
        - Enum Values: Translates between the values used locally by the actor to the symbol
        sent in messages.
        - Removes any key-value pairs where the value is None for a clearer message, especially
        in cases with many optional attributes.

        It also applies these changes recursively to sub-types.
        """
        d = {
            key: value
            for key, value in self.dict(
                by_alias=True, include=self.__fields_set__ | {"type_name", "version"}
            ).items()
            if value is not None
        }
        d["SupervisorContainer"] = self.supervisor_container.as_dict()
        # Recursively calling as_dict()
        gni_list = []
        for elt in self.gni_list:
            gni_list.append(elt.as_dict())
        d["GniList"] = gni_list
        return d

    def as_type(self) -> bytes:
        """
        Serialize to the super.starter.000 representation.

        Instances in the class are python-native representations of super.starter.000
        objects, while the actual super.starter.000 object is the serialized UTF-8 byte
        string designed for sending in a message.

        This method calls the as_dict() method, which differs from the native python dict()
        in the following key ways:
        - Enum Values: Translates between the values used locally by the actor to the symbol
        sent in messages.
        - - Removes any key-value pairs where the value is None for a clearer message, especially
        in cases with many optional attributes.

        It also applies these changes recursively to sub-types.

        Its near-inverse is SuperStarter.type_to_tuple(). If the type (or any sub-types)
        includes an enum, then the type_to_tuple will map an unrecognized symbol to the
        default enum value. This is why these two methods are only 'near' inverses.
        """
        json_string = json.dumps(self.as_dict())
        return json_string.encode("utf-8")

    def __hash__(self):
        return hash((type(self),) + tuple(self.__dict__.values()))  # noqa


class SuperStarter_Maker:
    type_name = "super.starter"
    version = "000"

    @classmethod
    def tuple_to_type(cls, tuple: SuperStarter) -> bytes:
        """
        Given a Python class object, returns the serialized JSON type object.
        """
        return tuple.as_type()

    @classmethod
    def type_to_tuple(cls, t: bytes) -> SuperStarter:
        """
        Given a serialized JSON type object, returns the Python class object.
        """
        try:
            d = json.loads(t)
        except TypeError:
            raise GwTypeError("Type must be string or bytes!")
        if not isinstance(d, dict):
            raise GwTypeError(f"Deserializing <{t}> must result in dict!")
        return cls.dict_to_tuple(d)

    @classmethod
    def dict_to_tuple(cls, d: dict[str, Any]) -> SuperStarter:
        """
        Deserialize a dictionary representation of a super.starter.000 message object
        into a SuperStarter python object for internal use.

        This is the near-inverse of the SuperStarter.as_dict() method:
          - Enums: translates between the symbols sent in messages between actors and
        the values used by the actors internally once they've deserialized the messages.
          - Types: recursively validates and deserializes sub-types.

        Note that if a required attribute with a default value is missing in a dict, this method will
        raise a GwTypeError. This differs from the pydantic BaseModel practice of auto-completing
        missing attributes with default values when they exist.

        Args:
            d (dict): the dictionary resulting from json.loads(t) for a serialized JSON type object t.

        Raises:
           GwTypeError: if the dict cannot be turned into a SuperStarter object.

        Returns:
            SuperStarter
        """
        d2 = dict(d)
        if "SupervisorContainer" not in d2.keys():
            raise GwTypeError(f"dict missing SupervisorContainer: <{d2}>")
        if not isinstance(d2["SupervisorContainer"], dict):
            raise GwTypeError(
                f"SupervisorContainer <{d2['SupervisorContainer']}> must be a SupervisorContainerGt!"
            )
        supervisor_container = SupervisorContainerGt_Maker.dict_to_tuple(
            d2["SupervisorContainer"]
        )
        d2["SupervisorContainer"] = supervisor_container
        if "GniList" not in d2.keys():
            raise GwTypeError(f"dict missing GniList: <{d2}>")
        if not isinstance(d2["GniList"], List):
            raise GwTypeError(f"GniList <{d2['GniList']}> must be a List!")
        gni_list = []
        for elt in d2["GniList"]:
            if not isinstance(elt, dict):
                raise GwTypeError(
                    f"GniList <{d2['GniList']}> must be a List of GNodeInstanceGt types"
                )
            t = GNodeInstanceGt_Maker.dict_to_tuple(elt)
            gni_list.append(t)
        d2["GniList"] = gni_list
        if "AliasWithKeyList" not in d2.keys():
            raise GwTypeError(f"dict missing AliasWithKeyList: <{d2}>")
        if "KeyList" not in d2.keys():
            raise GwTypeError(f"dict missing KeyList: <{d2}>")
        if "TypeName" not in d2.keys():
            raise GwTypeError(f"TypeName missing from dict <{d2}>")
        if "Version" not in d2.keys():
            raise GwTypeError(f"Version missing from dict <{d2}>")
        if d2["Version"] != "000":
            LOGGER.debug(
                f"Attempting to interpret super.starter version {d2['Version']} as version 000"
            )
            d2["Version"] = "000"
        return SuperStarter(**d2)


def check_is_left_right_dot(v: str) -> None:
    """Checks LeftRightDot Format

    LeftRightDot format: Lowercase alphanumeric words separated by periods, with
    the most significant word (on the left) starting with an alphabet character.

    Args:
        v (str): the candidate

    Raises:
        ValueError: if v is not LeftRightDot format
    """
    from typing import List

    try:
        x: List[str] = v.split(".")
    except:
        raise ValueError(f"Failed to seperate <{v}> into words with split'.'")
    first_word = x[0]
    first_char = first_word[0]
    if not first_char.isalpha():
        raise ValueError(
            f"Most significant word of <{v}> must start with alphabet char."
        )
    for word in x:
        if not word.isalnum():
            raise ValueError(f"words of <{v}> split by by '.' must be alphanumeric.")
    if not v.islower():
        raise ValueError(f"All characters of <{v}> must be lowercase.")



"""Type g.node.instance.gt, version 000"""

import json
import logging
from typing import Any
from typing import Dict
from typing import Literal
from typing import Optional

from gw.errors import DcError
from gw.errors import GwTypeError
from gw.utils import snake_to_camel
from pydantic import BaseModel
from pydantic import Extra
from pydantic import Field
from pydantic import validator

from gwbase.data_classes.g_node import GNode
from gwbase.data_classes.g_node_instance import GNodeInstance
from gwbase.enums import GniStatus
from gwbase.enums import StrategyName


LOG_FORMAT = (
    "%(levelname) -10s %(asctime)s %(name) -30s %(funcName) "
    "-35s %(lineno) -5d: %(message)s"
)
LOGGER = logging.getLogger(__name__)


class GNodeInstanceGt(BaseModel):
    """
    Used to send and receive updates about GNodeInstances.

    One of the layers of abstraction connecting a GNode with a running app in a Docker container.

    [More info](https://gridworks.readthedocs.io/en/latest/g-node-instance.html)
    """

    g_node_instance_id: str = Field(
        title="Immutable identifier for GNodeInstance (Gni)",
    )
    g_node_id: str = Field(
        title="The GNode represented by the Gni",
    )
    strategy: StrategyName = Field(
        title="Used to determine the code running in a GNode actor application",
    )
    status: GniStatus = Field(
        title="Lifecycle Status for Gni",
    )
    supervisor_container_id: str = Field(
        title="The Id of the docker container where the Gni runs",
    )
    start_time_unix_s: int = Field(
        title="When the gni starts representing the GNode",
        description=(
            "Specifically, when the Status changes from Pending to Active. Note that this is "
            "time in the GNode's World, which may not be real time if it is a simulation."
        ),
    )
    end_time_unix_s: int = Field(
        title="When the gni stops representing the GNode",
        description="Specifically, when the Status changes from Active to Done.",
    )
    algo_address: Optional[str] = Field(
        title="Algorand address for Gni",
        default=None,
    )
    type_name: Literal["g.node.instance.gt"] = "g.node.instance.gt"
    version: Literal["000"] = "000"

    class Config:
        extra = Extra.allow
        allow_population_by_field_name = True
        alias_generator = snake_to_camel

    @validator("g_node_instance_id")
    def _check_g_node_instance_id(cls, v: str) -> str:
        try:
            check_is_uuid_canonical_textual(v)
        except ValueError as e:
            raise ValueError(
                f"GNodeInstanceId failed UuidCanonicalTextual format validation: {e}"
            )
        return v

    @validator("g_node_id")
    def _check_g_node_id(cls, v: str) -> str:
        try:
            check_is_uuid_canonical_textual(v)
        except ValueError as e:
            raise ValueError(
                f"GNodeId failed UuidCanonicalTextual format validation: {e}"
            )
        return v

    @validator("supervisor_container_id")
    def _check_supervisor_container_id(cls, v: str) -> str:
        try:
            check_is_uuid_canonical_textual(v)
        except ValueError as e:
            raise ValueError(
                f"SupervisorContainerId failed UuidCanonicalTextual format validation: {e}"
            )
        return v

    @validator("start_time_unix_s")
    def _check_start_time_unix_s(cls, v: int) -> int:
        try:
            check_is_reasonable_unix_time_s(v)
        except ValueError as e:
            raise ValueError(
                f"StartTimeUnixS failed ReasonableUnixTimeS format validation: {e}"
            )
        return v

    @validator("algo_address")
    def _check_algo_address(cls, v: Optional[str]) -> Optional[str]:
        if v is None:
            return v
        try:
            check_is_algo_address_string_format(v)
        except ValueError as e:
            raise ValueError(
                f"AlgoAddress failed AlgoAddressStringFormat format validation: {e}"
            )
        return v

    def as_dict(self) -> Dict[str, Any]:
        """
        Translate the object into a dictionary representation that can be serialized into a
        g.node.instance.gt.000 object.

        This method prepares the object for serialization by the as_type method, creating a
        dictionary with key-value pairs that follow the requirements for an instance of the
        g.node.instance.gt.000 type. Unlike the standard python dict method,
        it makes the following substantive changes:
        - Enum Values: Translates between the values used locally by the actor to the symbol
        sent in messages.
        - Removes any key-value pairs where the value is None for a clearer message, especially
        in cases with many optional attributes.

        It also applies these changes recursively to sub-types.
        """
        d = {
            key: value
            for key, value in self.dict(
                by_alias=True, include=self.__fields_set__ | {"type_name", "version"}
            ).items()
            if value is not None
        }
        del d["Strategy"]
        d["StrategyGtEnumSymbol"] = StrategyName.value_to_symbol(self.strategy)
        del d["Status"]
        d["StatusGtEnumSymbol"] = GniStatus.value_to_symbol(self.status)
        return d

    def as_type(self) -> bytes:
        """
        Serialize to the g.node.instance.gt.000 representation.

        Instances in the class are python-native representations of g.node.instance.gt.000
        objects, while the actual g.node.instance.gt.000 object is the serialized UTF-8 byte
        string designed for sending in a message.

        This method calls the as_dict() method, which differs from the native python dict()
        in the following key ways:
        - Enum Values: Translates between the values used locally by the actor to the symbol
        sent in messages.
        - - Removes any key-value pairs where the value is None for a clearer message, especially
        in cases with many optional attributes.

        It also applies these changes recursively to sub-types.

        Its near-inverse is GNodeInstanceGt.type_to_tuple(). If the type (or any sub-types)
        includes an enum, then the type_to_tuple will map an unrecognized symbol to the
        default enum value. This is why these two methods are only 'near' inverses.
        """
        json_string = json.dumps(self.as_dict())
        return json_string.encode("utf-8")

    def __hash__(self):
        return hash((type(self),) + tuple(self.__dict__.values()))  # noqa


class GNodeInstanceGt_Maker:
    type_name = "g.node.instance.gt"
    version = "000"

    @classmethod
    def tuple_to_type(cls, tuple: GNodeInstanceGt) -> bytes:
        """
        Given a Python class object, returns the serialized JSON type object.
        """
        return tuple.as_type()

    @classmethod
    def type_to_tuple(cls, t: bytes) -> GNodeInstanceGt:
        """
        Given a serialized JSON type object, returns the Python class object.
        """
        try:
            d = json.loads(t)
        except TypeError:
            raise GwTypeError("Type must be string or bytes!")
        if not isinstance(d, dict):
            raise GwTypeError(f"Deserializing <{t}> must result in dict!")
        return cls.dict_to_tuple(d)

    @classmethod
    def dict_to_tuple(cls, d: dict[str, Any]) -> GNodeInstanceGt:
        """
        Deserialize a dictionary representation of a g.node.instance.gt.000 message object
        into a GNodeInstanceGt python object for internal use.

        This is the near-inverse of the GNodeInstanceGt.as_dict() method:
          - Enums: translates between the symbols sent in messages between actors and
        the values used by the actors internally once they've deserialized the messages.
          - Types: recursively validates and deserializes sub-types.

        Note that if a required attribute with a default value is missing in a dict, this method will
        raise a GwTypeError. This differs from the pydantic BaseModel practice of auto-completing
        missing attributes with default values when they exist.

        Args:
            d (dict): the dictionary resulting from json.loads(t) for a serialized JSON type object t.

        Raises:
           GwTypeError: if the dict cannot be turned into a GNodeInstanceGt object.

        Returns:
            GNodeInstanceGt
        """
        d2 = dict(d)
        if "GNodeInstanceId" not in d2.keys():
            raise GwTypeError(f"dict missing GNodeInstanceId: <{d2}>")
        if "GNodeId" not in d2.keys():
            raise GwTypeError(f"dict missing GNode: <{d2}>")
        if "StrategyGtEnumSymbol" not in d2.keys():
            raise GwTypeError(f"StrategyGtEnumSymbol missing from dict <{d2}>")
        value = StrategyName.symbol_to_value(d2["StrategyGtEnumSymbol"])
        d2["Strategy"] = StrategyName(value)
        del d2["StrategyGtEnumSymbol"]
        if "StatusGtEnumSymbol" not in d2.keys():
            raise GwTypeError(f"StatusGtEnumSymbol missing from dict <{d2}>")
        value = GniStatus.symbol_to_value(d2["StatusGtEnumSymbol"])
        d2["Status"] = GniStatus(value)
        del d2["StatusGtEnumSymbol"]
        if "SupervisorContainerId" not in d2.keys():
            raise GwTypeError(f"dict missing SupervisorContainerId: <{d2}>")
        if "StartTimeUnixS" not in d2.keys():
            raise GwTypeError(f"dict missing StartTimeUnixS: <{d2}>")
        if "EndTimeUnixS" not in d2.keys():
            raise GwTypeError(f"dict missing EndTimeUnixS: <{d2}>")
        if "TypeName" not in d2.keys():
            raise GwTypeError(f"TypeName missing from dict <{d2}>")
        if "Version" not in d2.keys():
            raise GwTypeError(f"Version missing from dict <{d2}>")
        if d2["Version"] != "000":
            LOGGER.debug(
                f"Attempting to interpret g.node.instance.gt version {d2['Version']} as version 000"
            )
            d2["Version"] = "000"
        return GNodeInstanceGt(**d2)

    @classmethod
    def tuple_to_dc(cls, t: GNodeInstanceGt) -> GNodeInstance:
        if t.g_node_instance_id in GNodeInstance.by_id.keys():
            dc = GNodeInstance.by_id[t.g_node_instance_id]
        else:
            try:
                g_node = GNode.by_id[t.g_node_id]
            except:
                raise DcError(f"Missing GNode with id {t.g_node_id}")
            dc = GNodeInstance(
                g_node_instance_id=t.g_node_instance_id,
                g_node=g_node,
                strategy=t.strategy,
                status=t.status,
                supervisor_container_id=t.supervisor_container_id,
                start_time_unix_s=t.start_time_unix_s,
                end_time_unix_s=t.end_time_unix_s,
                algo_address=t.algo_address,
            )
        return dc

    @classmethod
    def dc_to_tuple(cls, dc: GNodeInstance) -> GNodeInstanceGt:
        return GNodeInstanceGt(
            g_node_instance_id=dc.g_node_instance_id,
            g_node_id=dc.g_node_id,
            strategy=dc.strategy,
            status=dc.status,
            supervisor_container_id=dc.supervisor_container_id,
            start_time_unix_s=dc.start_time_unix_s,
            end_time_unix_s=dc.end_time_unix_s,
            algo_address=dc.algo_address,
        )

    @classmethod
    def type_to_dc(cls, t: str) -> GNodeInstance:
        return cls.tuple_to_dc(cls.type_to_tuple(t))

    @classmethod
    def dc_to_type(cls, dc: GNodeInstance) -> str:
        return cls.dc_to_tuple(dc).as_type()

    @classmethod
    def dict_to_dc(cls, d: dict[Any, str]) -> GNodeInstance:
        return cls.tuple_to_dc(cls.dict_to_tuple(d))


def check_is_algo_address_string_format(v: str) -> None:
    """
    AlgoAddressStringFormat format: The public key of a private/public Ed25519
    key pair, transformed into an  Algorand address, by adding a 4-byte checksum
    to the end of the public key and then encoding in base32.

    Raises:
        ValueError: if not AlgoAddressStringFormat format
    """
    import algosdk

    at = algosdk.abi.AddressType()
    try:
        result = at.decode(at.encode(v))
    except Exception as e:
        raise ValueError(f"Not AlgoAddressStringFormat: {e}")


def check_is_reasonable_unix_time_s(v: int) -> None:
    """Checks ReasonableUnixTimeS format

    ReasonableUnixTimeS format: unix seconds between Jan 1 2000 and Jan 1 3000

    Args:
        v (int): the candidate

    Raises:
        ValueError: if v is not ReasonableUnixTimeS format
    """
    from datetime import datetime
    from datetime import timezone

    start_date = datetime(2000, 1, 1, tzinfo=timezone.utc)
    end_date = datetime(3000, 1, 1, tzinfo=timezone.utc)

    start_timestamp = int(start_date.timestamp())
    end_timestamp = int(end_date.timestamp())

    if v < start_timestamp:
        raise ValueError(f"{v} must be after Jan 1 2000")
    if v > end_timestamp:
        raise ValueError(f"{v} must be before Jan 1 3000")


def check_is_uuid_canonical_textual(v: str) -> None:
    """Checks UuidCanonicalTextual format

    UuidCanonicalTextual format:  A string of hex words separated by hyphens
    of length 8-4-4-4-12.

    Args:
        v (str): the candidate

    Raises:
        ValueError: if v is not UuidCanonicalTextual format
    """
    try:
        x = v.split("-")
    except AttributeError as e:
        raise ValueError(f"Failed to split on -: {e}")
    if len(x) != 5:
        raise ValueError(f"<{v}> split by '-' did not have 5 words")
    for hex_word in x:
        try:
            int(hex_word, 16)
        except ValueError:
            raise ValueError(f"Words of <{v}> are not all hex")
    if len(x[0]) != 8:
        raise ValueError(f"<{v}> word lengths not 8-4-4-4-12")
    if len(x[1]) != 4:
        raise ValueError(f"<{v}> word lengths not 8-4-4-4-12")
    if len(x[2]) != 4:
        raise ValueError(f"<{v}> word lengths not 8-4-4-4-12")
    if len(x[3]) != 4:
        raise ValueError(f"<{v}> word lengths not 8-4-4-4-12")
    if len(x[4]) != 12:
        raise ValueError(f"<{v}> word lengths not 8-4-4-4-12")



from enum import auto
from typing import List
from typing import Optional

from gw.enums import GwStrEnum


class GNodeStatus(GwStrEnum):
    """
    Enum for managing GNode lifecycle

    Enum g.node.status version 100 in the GridWorks Type registry.

    Used by used by multiple Application Shared Languages (ASLs). For more information:
      - [ASLs](https://gridworks-type-registry.readthedocs.io/en/latest/)
      - [Global Authority](https://gridworks-type-registry.readthedocs.io/en/latest/enums.html#gnodestatus)
      - [More Info](https://gridworks.readthedocs.io/en/latest/g-node-status.html)

    Values (with symbols in parens):
      - Unknown (00000000): Default value
      - Pending (153d3475): The GNode exists but cannot be used yet.
      - Active (a2cfc2f7): The GNode can be used.
      - PermanentlyDeactivated (839b38db): The GNode can no longer be used, now or in the future.
      - Suspended (f5831e1d): The GNode cannot be used, but may become active in the future.
    """

    Unknown = auto()
    Pending = auto()
    Active = auto()
    PermanentlyDeactivated = auto()
    Suspended = auto()

    @classmethod
    def default(cls) -> "GNodeStatus":
        """
        Returns default value (in this case Unknown)
        """
        return cls.Unknown

    @classmethod
    def values(cls) -> List[str]:
        """
        Returns enum choices
        """
        return [elt.value for elt in cls]

    @classmethod
    def version(cls, value: Optional[str] = None) -> str:
        """
        Returns the version of the class (default) used by this package or the
        version of a candidate enum value (always less than or equal to the version
        of the class)

        Args:
            value (Optional[str]): None (for version of the Enum itself) or
            the candidate enum value.

        Raises:
            ValueError: If the value is not one of the enum values.

        Returns:
            str: The version of the enum used by this code (if given no
            value) OR the earliest version of the enum containing the value.
        """
        if value is None:
            return "100"
        if not isinstance(value, str):
            raise ValueError(f"This method applies to strings, not enums")
        if value not in value_to_version.keys():
            raise ValueError(f"Unknown enum value: {value}")
        return value_to_version[value]

    @classmethod
    def enum_name(cls) -> str:
        """
        The name in the GridWorks Type Registry (g.node.status)
        """
        return "g.node.status"

    @classmethod
    def enum_version(cls) -> str:
        """
        The version in the GridWorks Type Registry (100)
        """
        return "100"

    @classmethod
    def symbol_to_value(cls, symbol: str) -> str:
        """
        Given the symbol sent in a serialized message, returns the encoded enum.

        Args:
            symbol (str): The candidate symbol.

        Returns:
            str: The encoded value associated to that symbol. If the symbol is not
            recognized - which could happen if the actor making the symbol is using
            a later version of this enum, returns the default value of "Unknown".
        """
        if symbol not in symbol_to_value.keys():
            return cls.default().value
        return symbol_to_value[symbol]

    @classmethod
    def value_to_symbol(cls, value: str) -> str:
        """
        Provides the encoding symbol for a GNodeStatus enum to send in seriliazed messages.

        Args:
            symbol (str): The candidate value.

        Returns:
            str: The symbol encoding that value. If the value is not recognized -
            which could happen if the actor making the message used a later version
            of this enum than the actor decoding the message, returns the default
            symbol of "00000000".
        """
        if value not in value_to_symbol.keys():
            return value_to_symbol[cls.default().value]
        return value_to_symbol[value]

    @classmethod
    def symbols(cls) -> List[str]:
        """
        Returns a list of the enum symbols
        """
        return [
            "00000000",
            "153d3475",
            "a2cfc2f7",
            "839b38db",
            "f5831e1d",
        ]


symbol_to_value = {
    "00000000": "Unknown",
    "153d3475": "Pending",
    "a2cfc2f7": "Active",
    "839b38db": "PermanentlyDeactivated",
    "f5831e1d": "Suspended",
}

value_to_symbol = {value: key for key, value in symbol_to_value.items()}

value_to_version = {
    "Unknown": "100",
    "Pending": "100",
    "Active": "100",
    "PermanentlyDeactivated": "100",
    "Suspended": "100",
}



from enum import auto
from typing import List
from typing import Optional

from gw.enums import GwStrEnum


class StrategyName(GwStrEnum):
    """
    Used to assign code to run a particular GNodeInstance

    Enum strategy.name version 001 in the GridWorks Type registry.

    Used by used by multiple Application Shared Languages (ASLs). For more information:
      - [ASLs](https://gridworks-type-registry.readthedocs.io/en/latest/)
      - [Global Authority](https://gridworks-type-registry.readthedocs.io/en/latest/enums.html#strategyname)

    Values (with symbols in parens):
      - NoActor (00000000): Assigned to GNodes that do not have actors
      - WorldA (642c83d3): Authority on GNodeInstances, and their private keys. Maintains a FastAPI
        used for relational information about backoffice information held locally and/or in
        the GNodeRegistry/GNodeFactory. [More Info](https://gridworks.readthedocs.io/en/latest/world-role.html)
      - SupervisorA (4bb2cf7e): A simple supervisor that monitors its supervised AtomicTNode GNode
        strategies via a heartbeat health check. [More Info](https://gridworks.readthedocs.io/en/latest/supervisor.html)
      - AtnHeatPumpWithBoostStore (f5961401): AtomicTNode for a heat pump thermal storage heating
        system with a boost element and a thermal \n heated by the boost element. [More on AtomicTNodes](https://gridworks.readthedocs.io/en/latest/atomic-t-node.html)
      - TcGlobalA (73fbe6ab): Used to manage the global time of the Gridworks system when run with\n
        in a fully simulated universe. \n [More on TimeCoordinators](https://gridworks.readthedocs.io/en/latest/time-coordinator.html)
      - MarketMakerA (5e18a52e): Runs a two-sided market associated to its GNode as part of the copper
        GNode sub-tree. [More on MarketMakers](https://gridworks.readthedocs.io/en/latest/market-maker.html)
      - AtnBrickStorageHeater (b2a125d6): Publicly available Dijkstra-based AtomicTNode strategy
        for a brick storage heater. These heaters are rooom units that store heat in a brick
        core, are heated with resistive elements, and typically have a fan to blow air over
        the brick core. They are sometimes called Electric Thermal Storage (ETS) heaters, and
        in the UK are often called Economy 7 heaters or Night Storage Heaters. A strategy very
        similar to this was used by VCharge to manage an ETS fleet of several thousand heaters
        in Pennsylvania. This strategy is meant to serve as a template for other private strategies,
        and also allows for an end-to-end simulation of a realistic aggregated transactive load
        capable of generating a highly elastic bid curve [More Info](https://gridworks-atn.readthedocs.io/en/latest/brick-storage-heater.html)
    """

    NoActor = auto()
    WorldA = auto()
    SupervisorA = auto()
    AtnHeatPumpWithBoostStore = auto()
    TcGlobalA = auto()
    MarketMakerA = auto()
    AtnBrickStorageHeater = auto()

    @classmethod
    def default(cls) -> "StrategyName":
        """
        Returns default value (in this case NoActor)
        """
        return cls.NoActor

    @classmethod
    def values(cls) -> List[str]:
        """
        Returns enum choices
        """
        return [elt.value for elt in cls]

    @classmethod
    def version(cls, value: Optional[str] = None) -> str:
        """
        Returns the version of the class (default) used by this package or the
        version of a candidate enum value (always less than or equal to the version
        of the class)

        Args:
            value (Optional[str]): None (for version of the Enum itself) or
            the candidate enum value.

        Raises:
            ValueError: If the value is not one of the enum values.

        Returns:
            str: The version of the enum used by this code (if given no
            value) OR the earliest version of the enum containing the value.
        """
        if value is None:
            return "001"
        if not isinstance(value, str):
            raise ValueError(f"This method applies to strings, not enums")
        if value not in value_to_version.keys():
            raise ValueError(f"Unknown enum value: {value}")
        return value_to_version[value]

    @classmethod
    def enum_name(cls) -> str:
        """
        The name in the GridWorks Type Registry (strategy.name)
        """
        return "strategy.name"

    @classmethod
    def enum_version(cls) -> str:
        """
        The version in the GridWorks Type Registry (001)
        """
        return "001"

    @classmethod
    def symbol_to_value(cls, symbol: str) -> str:
        """
        Given the symbol sent in a serialized message, returns the encoded enum.

        Args:
            symbol (str): The candidate symbol.

        Returns:
            str: The encoded value associated to that symbol. If the symbol is not
            recognized - which could happen if the actor making the symbol is using
            a later version of this enum, returns the default value of "NoActor".
        """
        if symbol not in symbol_to_value.keys():
            return cls.default().value
        return symbol_to_value[symbol]

    @classmethod
    def value_to_symbol(cls, value: str) -> str:
        """
        Provides the encoding symbol for a StrategyName enum to send in seriliazed messages.

        Args:
            symbol (str): The candidate value.

        Returns:
            str: The symbol encoding that value. If the value is not recognized -
            which could happen if the actor making the message used a later version
            of this enum than the actor decoding the message, returns the default
            symbol of "00000000".
        """
        if value not in value_to_symbol.keys():
            return value_to_symbol[cls.default().value]
        return value_to_symbol[value]

    @classmethod
    def symbols(cls) -> List[str]:
        """
        Returns a list of the enum symbols
        """
        return [
            "00000000",
            "642c83d3",
            "4bb2cf7e",
            "f5961401",
            "73fbe6ab",
            "5e18a52e",
            "b2a125d6",
        ]


symbol_to_value = {
    "00000000": "NoActor",
    "642c83d3": "WorldA",
    "4bb2cf7e": "SupervisorA",
    "f5961401": "AtnHeatPumpWithBoostStore",
    "73fbe6ab": "TcGlobalA",
    "5e18a52e": "MarketMakerA",
    "b2a125d6": "AtnBrickStorageHeater",
}

value_to_symbol = {value: key for key, value in symbol_to_value.items()}

value_to_version = {
    "NoActor": "000",
    "WorldA": "000",
    "SupervisorA": "000",
    "AtnHeatPumpWithBoostStore": "000",
    "TcGlobalA": "000",
    "MarketMakerA": "000",
    "AtnBrickStorageHeater": "001",
}



import datetime
import enum
import functools
import logging
import random
import threading
import time
import uuid
from abc import ABC
from abc import abstractmethod
from enum import auto
from typing import Dict
from typing import List
from typing import Optional
from typing import no_type_check

import gw.utils as utils
import pika
from gw.enums import GwStrEnum
from gw.errors import GwTypeError
from pika.channel import Channel as PikaChannel
from pika.spec import Basic
from pika.spec import BasicProperties

import gwbase.types.base_asl_types as base_asl_types
from gwbase.config import GNodeSettings
from gwbase.enums import GNodeRole
from gwbase.enums import MessageCategory
from gwbase.enums import MessageCategorySymbol
from gwbase.enums import UniverseType
from gwbase.types import HeartbeatA
from gwbase.types import HeartbeatA_Maker
from gwbase.types import SimTimestep
from gwbase.types import SimTimestep_Maker


class RabbitRole(GwStrEnum):
    atomictnode = auto()
    gnode = auto()
    marketmaker = auto()
    scada = auto()
    supervisor = auto()
    timecoordinator = auto()
    world = auto()

    @classmethod
    def values(cls) -> List[str]:
        return [elt.value for elt in cls]


RoleByRabbitRole: Dict[RabbitRole, GNodeRole] = {
    RabbitRole.atomictnode: GNodeRole.AtomicTNode,
    RabbitRole.gnode: GNodeRole.GNode,
    RabbitRole.marketmaker: GNodeRole.MarketMaker,
    RabbitRole.scada: GNodeRole.Scada,
    RabbitRole.supervisor: GNodeRole.Supervisor,
    RabbitRole.timecoordinator: GNodeRole.TimeCoordinator,
    RabbitRole.world: GNodeRole.World,
}


RabbitRolebyRole: Dict[GNodeRole, RabbitRole] = {
    GNodeRole.AtomicTNode: RabbitRole.atomictnode,
    GNodeRole.GNode: RabbitRole.gnode,
    GNodeRole.MarketMaker: RabbitRole.marketmaker,
    GNodeRole.Scada: RabbitRole.scada,
    GNodeRole.Supervisor: RabbitRole.supervisor,
    GNodeRole.TimeCoordinator: RabbitRole.timecoordinator,
    GNodeRole.World: RabbitRole.world,
}


class OnSendMessageDiagnostic(enum.Enum):
    CHANNEL_NOT_OPEN = "ChannelNotOpen"
    STOPPED_SO_NOT_SENDING = "StoppedSoNotSending"
    STOPPING_SO_NOT_SENDING = "StoppingSoNotSending"
    MESSAGE_SENT = "MessageSent"
    UNKNOWN_ERROR = "UnknownError"


class OnReceiveMessageDiagnostic(enum.Enum):
    TYPE_NAME_DECODING_PROBLEM = "TypeNameDecodingProblem"
    UNKNOWN_ROUTING_KEY_TYPE = "UnknownRoutingKeyType"
    UNHANDLED_ROUTING_KEY_TYPE = "UnhandledRoutingKeyType"
    UNKNOWN_MESSAGE_CATEGORY_SYMBOL = "UnknownMessageCategorySymbol"
    UNKNOWN_TYPE_NAME = "UnknownTypeName"
    FROM_GNODE_DECODING_PROBLEM = "FromGNodeDecodingProblem"
    UNKONWN_GNODE = "UnknownGNode"
    TO_DIRECT_ROUTING = "ToDirectRouting"


BACKOFF_NUMBER = 16

LOG_FORMAT = (
    "%(levelname) -10s %(asctime)s %(name) -30s %(funcName) "
    "-35s %(lineno) -5d: %(message)s"
)
LOGGER = logging.getLogger(__name__)

LOGGER.setLevel(logging.INFO)


class ActorBase(ABC):
    "This is the base class for GNodes, used to communicate via RabbitMQ"
    _url: str

    SHUTDOWN_INTERVAL: float = 0.1

    def __init__(
        self,
        settings: GNodeSettings,
        api_type_maker_by_name: Dict[
            str, HeartbeatA_Maker
        ] = base_asl_types.TypeMakerByName,
    ):
        self.settings: GNodeSettings = settings
        self.api_type_maker_by_name = api_type_maker_by_name
        self.latest_routing_key: Optional[str] = None
        self.shutting_down: bool = False
        self.alias: str = settings.g_node_alias
        self.g_node_instance_id: str = settings.g_node_instance_id
        self.g_node_role: GNodeRole = GNodeRole(settings.g_node_role_value)
        self.rabbit_role: RabbitRole = RabbitRolebyRole[self.g_node_role]
        self.universe_type: UniverseType = UniverseType(settings.universe_type_value)
        # Used for tracking time in simulated worlds
        self._time: float = time.time()
        if self.universe_type == UniverseType.Dev:
            self._time = settings.initial_time_unix_s
        self.actor_main_stopped: bool = True

        adder = "-F" + str(uuid.uuid4()).split("-")[0][0:3]
        self.queue_name: str = self.alias + adder
        self._consume_exchange: str = self.rabbit_role.value + "_tx"
        self._publish_exchange: str = self.rabbit_role.value + "mic_tx"

        self._consume_connection: Optional[
            pika.adapters.select_connection.SelectConnection
        ] = None
        self._consume_channel: Optional[pika.channel.Channel] = None
        self._closing_consumer: bool = False
        self._consumer_tag: Optional[str] = None
        self.should_reconnect_consumer: bool = False
        self.was_consuming: bool = False
        self._consuming: bool = False
        # In production, experiment with higher prefetch values
        # for higher consumer throughput
        self._prefetch_count: int = 1
        self._reconnect_delay: int = 0
        self._url: str = settings.rabbit.url.get_secret_value()

        self.is_debug_mode: bool = False
        self.consuming_thread: threading.Thread = threading.Thread(
            target=self.run_reconnecting_consumer
        )
        self.publishing_thread: threading.Thread = threading.Thread(
            target=self.run_publisher
        )
        self._publish_connection: Optional[
            pika.adapters.select_connection.SelectConnection
        ] = None
        self._publish_channel: Optional[PikaChannel] = None
        self._stopping: bool = False
        self._stopped: bool = True
        self._latest_on_message_diagnostic: Optional[OnReceiveMessageDiagnostic] = None

    def start(self) -> None:
        self.local_start()
        self._stopped = False
        self.consuming_thread.start()
        # self.publishing_thread.start()

    def stop(self) -> None:
        self.shutting_down = True
        self.prepare_for_death()
        while self.actor_main_stopped is False:
            time.sleep(self.SHUTDOWN_INTERVAL)
        # self.stop_publisher()
        self.stop_consumer()
        self.local_stop()
        self.consuming_thread.join()
        # self.publishing_thread.join()
        self._stopping = False
        self._stopped = True

    def local_start(self) -> None:
        """This should be overwritten in derived class for additional threads.
        It cannot assume the rabbit channels are established and that
        messages can be received or sent."""
        self.actor_main_stopped = False

    def local_rabbit_startup(self) -> None:
        """This should be overwritten in derived class for any additional rabbit
        bindings. DO NOT start queues here.  It is called at the end of
        self.start_consuming()"""
        pass

    @abstractmethod
    def prepare_for_death(self) -> None:
        """Use actor_main_stopped to exit out of any threads in the derived class. Then
        use local_stop to join those threads.

        If there are no threads in the derived class, copy this method into the derived
        class, get rid of the abstractmethod decorator, and delete the exception"""
        self.actor_main_stopped = True
        raise NotImplementedError

    def local_stop(self) -> None:
        """Join any threads in the derived class."""
        pass

    def __repr__(self) -> str:
        return f"{self.alias}"

    ########################
    # Core Rabbit infrastructure
    ########################

    def flush_consumer(self) -> None:
        self.should_reconnect_consumer = False
        self.was_consuming = False
        self._consume_connection = None
        self._consume_channel = None
        self._closing_consumer = False
        self._consumer_tag = None
        self._consuming = False

    def run_reconnecting_consumer(self) -> None:
        while not self.actor_main_stopped:
            self.run_consumer()
            self._maybe_reconnect_consumer()

    def _maybe_reconnect_consumer(self) -> None:
        if self.should_reconnect_consumer:
            self.stop_consumer()
            reconnect_delay = self._get_reconnect_delay()
            if not self.actor_main_stopped:
                LOGGER.info("Reconnecting after %d seconds", reconnect_delay)
            time.sleep(reconnect_delay)
            self.flush_consumer()

    def _get_reconnect_delay(self) -> int:
        if self.was_consuming:
            self._reconnect_delay = 0
        else:
            self._reconnect_delay += 1
        if self._reconnect_delay > 30:
            self._reconnect_delay = 30
        return self._reconnect_delay

    def connect_consumer(self) -> pika.SelectConnection:
        """This method connects to RabbitMQ, returning the connection handle.
        When the connection is established, the on_consumer_connection_open method
        will be invoked by pika.
        :rtype: pika.SelectConnection
        """
        LOGGER.info("Connecting to %s", self._url)
        return pika.SelectConnection(
            parameters=pika.URLParameters(self._url),
            on_open_callback=self.on_consumer_connection_open,  # type: ignore[arg-type]
            on_open_error_callback=self.on_consumer_connection_open_error,  # type: ignore[arg-type]
            on_close_callback=self.on_consumer_connection_closed,  # type: ignore[arg-type]
        )

    def close_consumer_connection(self) -> None:
        self._consuming = False
        if self._consume_connection:
            if (
                not self._consume_connection.is_closing
                and not self._consume_connection.is_closed
            ):
                LOGGER.info("Closing consume connection")
                self._consume_connection.close()

    def on_consumer_connection_open(
        self, _unused_connection: pika.SelectConnection
    ) -> None:
        """This method is called by pika once the connection to RabbitMQ has
        been established. It passes the handle to the connection object in
        case we need it, but in this case, we'll just mark it unused.
        :param pika.SelectConnection _unused_connection: The connection
        """
        LOGGER.info("Connection opened")
        self.open_consume_channel()

    def on_consumer_connection_open_error(
        self, _unused_connection: pika.SelectConnection, err: Exception
    ) -> None:
        """This method is called by pika if the connection to RabbitMQ
        can't be established.
        :param pika.SelectConnection _unused_connection: The connection
        :param Exception err: The error
        """
        LOGGER.error(f"Consumer connection open failed: {err}")
        self.reconnect_consumer()

    def on_consumer_connection_closed(
        self, _unused_connection: pika.SelectConnection, reason: Exception
    ) -> None:
        """This method is invoked by pika when the connection to RabbitMQ is
        closed unexpectedly. Since it is unexpected, we will reconnect to
        RabbitMQ if it disconnects.
        :param pika.connection.Connection connection: The closed connection obj
        :param Exception reason: exception representing reason for loss of
        connection.
        """
        self._consume_channel = None
        if self._closing_consumer:
            self._consume_connection.ioloop.stop()  # type: ignore[union-attr]
        else:
            LOGGER.warning(f"Consumer connection closed, reconnect necessary: {reason}")
            self.reconnect_consumer()

    def reconnect_consumer(self) -> None:
        """Will be invoked if the connection can't be opened or is
        closed. Indicates that a reconnect is necessary then stops the
        ioloop.
        """
        self.should_reconnect_consumer = True
        self.stop_consumer()

    def open_consume_channel(self) -> None:
        """Open a new channel with RabbitMQ by issuing the Channel.Open RPC
        command. When RabbitMQ responds that the channel is open, the
        on_channel_open callback will be invoked by pika.
        """
        LOGGER.info("Creating a new channel")
        self._consume_connection.channel(on_open_callback=self.on_consumer_channel_open)  # type: ignore

    @no_type_check
    def on_consumer_channel_open(self, channel: PikaChannel) -> None:
        """Invoked by pika when the channel has been successfully opened.

        This callback is triggered when the channel is opened, and it provides
        the channel object that can be used for further operations. In this case,
        we'll proceed to declare the exchange to be used.

        :param channel: The opened channel object.
        :type channel: PikaChannel
        """
        LOGGER.info("Channel opened")
        self._consume_channel = channel
        self.add_on_consume_channel_close_callback()
        self.setup_exchange()

    @no_type_check
    def add_on_consume_channel_close_callback(self) -> None:
        """This method tells pika to call the on_consumer_channel_closed method if
        RabbitMQ unexpectedly closes the channel.
        """
        LOGGER.info("Adding consumer channel close callback")
        self._consume_channel.add_on_close_callback(self.on_consumer_channel_closed)

    @no_type_check
    def on_consumer_channel_closed(
        self, channel: PikaChannel, reason: Exception
    ) -> None:
        """Invoked by pika when the RabbitMQ channel is unexpectedly closed.

        This callback is triggered when a channel is closed, usually due to
        violating the protocol by attempting to re-declare an exchange or queue
        with different parameters. In this case, the connection is closed to
        gracefully shutdown the object.

        :param channel: The closed channel object.
        :type channel: PikaChannel
        :param reason: why the channel was closed
        :type reason: Exception
        """
        LOGGER.warning("Consume channel %i was closed: %s", channel, reason)
        self.close_consumer_connection()

    @no_type_check
    def setup_exchange(self) -> None:
        """Setup the exchange on RabbitMQ by invoking the Exchange.Declare RPC
        command. When it is complete, the on_exchange_declareok method will
        be invoked by pika.
        :param str|unicode exchange_name: The name of the exchange to declare
        """
        LOGGER.info("Declaring exchange: %s", self._consume_exchange)
        # Note: using functools.partial is not required, it is demonstrating
        # how arbitrary data can be passed to the callback when it is called
        cb = functools.partial(
            self.on_exchange_declareok, userdata=self._consume_exchange
        )
        self._consume_channel.exchange_declare(
            exchange=self._consume_exchange,
            exchange_type="topic",
            durable=True,
            internal=True,
            callback=cb,
        )

    @no_type_check
    def on_exchange_declareok(self, _unused_frame, userdata) -> None:
        """Invoked by pika when RabbitMQ has finished the Exchange.Declare RPC
        command.
        :param pika.Frame.Method unused_frame: Exchange.DeclareOk response frame
        :param str|unicode userdata: Extra user data (exchange name)
        """
        LOGGER.info("Exchange declared: %s", userdata)
        self.setup_queue()

    @no_type_check
    def setup_queue(self) -> None:
        """Setup the queue on RabbitMQ by invoking the Queue.Declare RPC
        command. When it is complete, the on_queue_declareok method will
        be invoked by pika.
        :param str|unicode queue_name: The name of the queue to declare.
        """
        LOGGER.info(f"Declaring queue {self.queue_name}")
        cb = functools.partial(self.on_queue_declareok)
        self._consume_channel.queue_declare(
            queue=self.queue_name, auto_delete=True, callback=cb
        )

    @no_type_check
    def on_queue_declareok(self, _unused_frame) -> None:
        """Method invoked by pika when the Queue.Declare RPC call made in
        setup_queue has completed. In this method we will bind the queue
        and exchange together with the routing key by issuing the Queue.Bind
        RPC command. When this command is complete, the on_bindok method will
        be invoked by pika.
        :param pika.frame.Method _unused_frame: The Queue.DeclareOk frame
        :param str|unicode userdata: Extra user data (queue name)
        """
        lrh_alias = self.alias.replace(".", "-")
        rj = MessageCategorySymbol.rj.value
        direct_message_to_me_binding = f"{rj}.*.*.*.*.{lrh_alias}"

        LOGGER.info(
            "Binding %s to %s with %s",
            self._consume_exchange,
            self.queue_name,
            direct_message_to_me_binding,
        )
        cb = functools.partial(
            self.on_direct_message_bindok, binding=direct_message_to_me_binding
        )
        self._consume_channel.queue_bind(
            self.queue_name,
            self._consume_exchange,
            routing_key=direct_message_to_me_binding,
            callback=cb,
        )

    @no_type_check
    def on_direct_message_bindok(self, _unused_frame, binding) -> None:
        """Invoked by pika when the Queue.Bind method has completed for direct messages. At this
        point we will set the prefetch count for the channel.
        :param pika.frame.Method _unused_frame: The Queue.BindOk response frame
        :param str|unicode userdata: Extra user data (queue name)
        """
        LOGGER.info(f"Queue {self.queue_name} bound with {binding}")
        self.set_qos()

    @no_type_check
    def set_qos(self) -> None:
        """This method sets up the consumer prefetch to only be delivered
        one message at a time. The consumer must acknowledge this message
        before RabbitMQ will deliver another one. You should experiment
        with different prefetch values to achieve desired performance.
        """
        self._consume_channel.basic_qos(
            prefetch_count=self._prefetch_count, callback=self.on_basic_qos_ok
        )

    @no_type_check
    def on_basic_qos_ok(self, _unused_frame) -> None:
        """Invoked by pika when the Basic.QoS method has completed. At this
        point we will start consuming messages by calling start_consuming
        which will invoke the needed RPC commands to start the process.
        :param pika.frame.Method _unused_frame: The Basic.QosOk response frame
        """
        LOGGER.info("QOS set to: %d", self._prefetch_count)
        self.start_consuming()

    @no_type_check
    def start_consuming(self) -> None:
        """This method sets up the consumer by first calling
        add_on_cancel_consumer_callback so that the object is notified if RabbitMQ
        cancels the consumer. It then issues the Basic.Consume RPC command
        which returns the consumer tag that is used to uniquely identify the
        consumer with RabbitMQ. We keep the value to use it when we want to
        cancel consuming. The on_message method is passed in as a callback pika
        will invoke when a message is fully received.
        """
        LOGGER.info("Start consuming")
        self.add_on_cancel_consumer_callback()
        self._consumer_tag = self._consume_channel.basic_consume(
            self.queue_name, self.on_message
        )
        self.was_consuming = True
        self._consuming = True
        self.local_rabbit_startup()

    def add_on_cancel_consumer_callback(self) -> None:
        """Add a callback that will be invoked if RabbitMQ cancels the consumer
        for some reason. If RabbitMQ does cancel the consumer,
        on_consumer_cancelled will be invoked by pika.
        """
        LOGGER.info("Adding consumer cancellation callback")
        self._consume_channel.add_on_cancel_callback(self.on_consumer_cancelled)  # type: ignore

    @no_type_check
    def on_consumer_cancelled(self, method_frame) -> None:
        """Invoked by pika when RabbitMQ sends a Basic.Cancel for a consumer
        receiving messages.
        :param pika.frame.Method method_frame: The Basic.Cancel frame
        """
        LOGGER.info("Consumer was cancelled remotely, shutting down: %r", method_frame)
        if self._consume_channel:
            self._consume_channel.close()

    @no_type_check
    def acknowledge_message(self, delivery_tag) -> None:
        """Acknowledge the message delivery from RabbitMQ by sending a
        Basic.Ack RPC method for the delivery tag.
        :param int delivery_tag: The delivery tag from the Basic.Deliver frame
        """
        LOGGER.debug(
            f"Acknowledging message {delivery_tag}",
        )
        self._consume_channel.basic_ack(delivery_tag)

    def stop_consuming(self) -> None:
        """Tell RabbitMQ that you would like to stop consuming by sending the
        Basic.Cancel RPC command.
        """
        if self._consume_channel:
            LOGGER.info("Sending a Basic.Cancel RPC command to RabbitMQ")
            cb = functools.partial(
                self.on_cancelconsumer_ok, userdata=self._consumer_tag
            )
            self._consume_channel.basic_cancel(self._consumer_tag, cb)  # type: ignore[arg-type]

    @no_type_check
    def on_cancelconsumer_ok(self, _unused_frame, userdata) -> None:
        """This method is invoked by pika when RabbitMQ acknowledges the
        cancellation of a consumer. At this point we will close the channel.
        This will invoke the on_consumer_channel_closed method once the channel has been
        closed, which will in-turn close the connection.
        :param pika.frame.Method _unused_frame: The Basic.CancelOk frame
        :param str|unicode userdata: Extra user data (consumer tag)
        """
        self._consuming = False
        LOGGER.info(
            "RabbitMQ acknowledged the cancellation of the consumer: %s", userdata
        )
        self.close_consumer_channel()
        self._closing_consumer = False

    def close_consumer_channel(self) -> None:
        """Call to close the channel with RabbitMQ cleanly by issuing the
        Channel.Close RPC command.
        """
        if self._consume_channel:
            if (
                not self._consume_channel.is_closing
                and not self._consume_channel.is_closed
            ):
                self._consume_channel.close()

    def run_consumer(self) -> None:
        """Run the example consumer by connecting to RabbitMQ and then
        starting the IOLoop to block and allow the SelectConnection to operate.
        """

        self._consume_connection = self.connect_consumer()
        self._consume_connection.ioloop.start()

    def stop_consumer(self) -> None:
        """Cleanly shutdown the connection to RabbitMQ by stopping the consumer
        with RabbitMQ. When RabbitMQ confirms the cancellation, on_cancelconsumer_ok
        will be invoked by pika, which will then closing the channel and
        connection. If you want to use this with CTRL-C, figure out
        how to add back the commented out ioloop.start() below without error.
        """
        if not self._closing_consumer:
            self._closing_consumer = True
            LOGGER.info("Consumer connection stopping")
            if self._consuming:
                self.stop_consuming()
                # self._consume_connection.ioloop.start()
            else:
                self._consume_connection.ioloop.stop()  # type: ignore

            LOGGER.info("Consumer connection stopped")

    def connect_publisher(self) -> pika.SelectConnection:
        """This method connects to RabbitMQ, returning the connection handle.
        When the connection is established, the on_connection_open method
        will be invoked by pika.
        :rtype: pika.SelectConnection
        """
        LOGGER.info("Setting up publisher connection to %s", self._url)
        return pika.SelectConnection(
            pika.URLParameters(self._url),
            on_open_callback=self.on_publish_connection_open,
            on_open_error_callback=self.on_publish_connection_open_error,
            on_close_callback=self.on_publish_connection_closed,
        )

    @no_type_check
    def on_publish_connection_open(self, _unused_connection) -> None:
        """This method is called by pika once the publisher connection to RabbitMQ has
        been established. It passes the handle to the connection object in
        case we need it, but in this case, we'll just mark it unused.
        :param pika.SelectConnection _unused_connection: The connection
        """
        LOGGER.info("Producer connection opened")
        self.open_publish_channel()

    @no_type_check
    def on_publish_connection_open_error(self, _unused_connection, err) -> None:
        """This method is called by pika if the connection to RabbitMQ
        can't be established.
        :param pika.SelectConnection _unused_connection: The connection
        :param Exception err: The error
        """
        LOGGER.error("Producer connection open failed, reopening in 1 second: %s", err)
        self._publish_connection.ioloop.call_later(
            1, self._publish_connection.ioloop.stop
        )

    @no_type_check
    def on_publish_connection_closed(self, _unused_connection, reason) -> None:
        """This method is invoked by pika when the connection to RabbitMQ is
        closed unexpectedly. Since it is unexpected, we will reconnect to
        RabbitMQ if it disconnects.
        :param pika.connection.Connection connection: The closed connection obj
        :param Exception reason: exception representing reason for loss of
        connection.
        """
        self._publish_channel = None
        if self._stopping:
            self._publish_connection.ioloop.stop()
        else:
            LOGGER.warning("Connection closed, reopening in 1 second: %s", reason)
            self._publish_connection.ioloop.call_later(
                1, self._publish_connection.ioloop.stop
            )

    def open_publish_channel(self) -> None:
        """This method will open a new channel with RabbitMQ by issuing the
        Channel.Open RPC command. When RabbitMQ confirms the channel is open
        by sending the Channel.OpenOK RPC reply, the on_channel_open method
        will be invoked.
        """
        LOGGER.info("Creating a new publish channel")
        self._publish_connection.channel(on_open_callback=self.on_publish_channel_open)  # type: ignore

    @no_type_check
    def on_publish_channel_open(self, channel: PikaChannel) -> None:
        """Invoked by pika when the channel has been successfully opened.

        This callback is triggered when the channel is opened, and it provides
        the channel object that can be used for further operations. In this case,
        we'll proceed to declare the exchange to be used.

        :param channel: The opened channel object.
        :type channel: PikaChannel
        """
        LOGGER.info("Publish channel opened")
        self._publish_channel = channel
        self.add_on_publish_channel_close_callback()

    def add_on_publish_channel_close_callback(self) -> None:
        """This method tells pika to call the on_channel_closed method if
        RabbitMQ unexpectedly closes the channel.
        """
        LOGGER.info("Adding channel close callback")
        self._publish_channel.add_on_close_callback(self.on_publish_channel_closed)  # type: ignore

    @no_type_check
    def on_publish_channel_closed(
        self, channel: PikaChannel, reason: Exception
    ) -> None:
        """Invoked by pika when the RabbitMQ channel is unexpectedly closed.

        This callback is triggered when a channel is closed, usually due to
        violating the protocol by attempting to re-declare an exchange or queue
        with different parameters. In this case, the connection is closed to
        gracefully shutdown the object.

        :param channel: The closed channel object.
        :type channel: PikaChannel
        :param reason: The reason why the channel was closed.
        :type reason: Exception
        """
        LOGGER.warning(f"Publish channel {channel} was closed: {reason}")
        self._publish_channel = None
        if not self._stopping:
            self._publish_connection.close()

    def run_publisher(self) -> None:
        """Run the example code by connecting and then starting the IOLoop."""
        while not self._stopping:
            self._publish_connection = None
            try:
                self._publish_connection = self.connect_publisher()
                self._publish_connection.ioloop.start()
            except KeyboardInterrupt:
                self.stop_publisher()
                if (
                    self._publish_connection is not None
                    and not self._publish_connection.is_closed
                ):
                    # Finish closing
                    self._publish_connection.ioloop.start()

        LOGGER.info("Stopped")

    def stop_publisher(self) -> None:
        """Stop the example by closing the channel and connection. We
        set a flag here so that we stop scheduling new messages to be
        published. The IOLoop is started because this method is
        invoked by the Try/Catch below when KeyboardInterrupt is caught.
        Starting the IOLoop again will allow the publisher to cleanly
        disconnect from RabbitMQ.
        """
        LOGGER.info(
            "Stopping RabbitMq message production - closing channel and connection"
        )
        self._stopping = True
        self.close_publish_channel()
        self.close_publish_connection()

    def close_publish_channel(self) -> None:
        """Invoke this command to close the channel with RabbitMQ by sending
        the Channel.Close RPC command.
        """
        if self._publish_channel:
            if (
                not self._publish_channel.is_closing
                and not self._publish_channel.is_closed
            ):
                self._publish_channel.close()

    def close_publish_connection(self) -> None:
        """This method closes the production connection to RabbitMQ."""
        if self._publish_connection:
            if (
                not self._publish_connection.is_closing
                and not self._publish_connection.is_closed
            ):
                self._publish_connection.close()
                LOGGER.info("Closing publish connection")

    ########################
    # Message passing semantics
    ########################

    @no_type_check
    def get_payload_type_name(self, basic_deliver) -> str:
        """The TypeName is a string that provides the strongly typed specification
            (API/ABI) for the incoming message. This is similar to knowing
            the protobuf name/method or the ABI name/method.

            The TypeName will articulate, in particular, how
            to decode the payload.

        Args:
            basic_deliver: the rabbit basic_deliver object
            body: the rabbit body object (i.e. the payload as incoming type)

        Returns:
            str: raises GwTypeError if the TypeName is not accessible.
            Otherwise returns the TypeName

        """
        # TODO: right now we encode this in the routing key. However, we
        # could also add it to a different basic_deliver property, which
        # might be easier for developers to grock.

        try:
            type_name = self.type_name_from_routing_key(
                routing_key=basic_deliver.routing_key,
            )
        except GwTypeError as e:
            LOGGER.info(f"Could not figure out TypeName: {e}")
            raise GwTypeError(f"{e}")
        return type_name

    def broadcast_routing_key(
        self, payload: HeartbeatA, radio_channel: Optional[str]
    ) -> str:
        msg_type = MessageCategorySymbol.rjb.value
        from_alias_lrh = self.alias.replace(".", "-")
        type_name_lrh = payload.type_name.replace(".", "-")
        from_role = self.rabbit_role.value
        if radio_channel is None:
            return f"{msg_type}.{from_alias_lrh}.{from_role}.{type_name_lrh}"
        else:
            check_is_left_right_dot(radio_channel)
            return f"{msg_type}.{from_alias_lrh}.{from_role}.{type_name_lrh}.{radio_channel}"

    def scada_routing_key(self, payload: HeartbeatA, to_g_node_alias: str) -> str:
        if self.g_node_role != GNodeRole.AtomicTNode:
            raise Exception(f"Only send messages to SCADA if role is AtomicTNode!")
        msg_type = MessageCategorySymbol.gw.value
        from_alias_lrh = self.alias.replace(".", "-")
        type_name_lrh = payload.type_name.replace(".", "-")

        scada_routing_key = f"{msg_type}.{from_alias_lrh}.{type_name_lrh}"
        return scada_routing_key

    def direct_routing_key(
        self, to_role: GNodeRole, payload: HeartbeatA, to_g_node_alias: str
    ) -> str:
        msg_type = MessageCategorySymbol.rj.value
        from_lrh_alias = self.alias.replace(".", "-")
        from_role = self.rabbit_role.value
        to_role_val = RabbitRolebyRole[to_role].value
        to_lrh_alias = to_g_node_alias.replace(".", "-")
        type_name_lrh = payload.type_name.replace(".", "-")

        direct_routing_key = f"{msg_type}.{from_lrh_alias}.{from_role}.{type_name_lrh}.{to_role_val}.{to_lrh_alias}"
        return direct_routing_key

    def message_category_from_routing_key(self, routing_key: str) -> MessageCategory:
        """Returns the MessageCategory of the message given the routing key.

        Raises a GwTypeError exception if there is a problem decoding the MessageCategory

        Args:
           routing_key (str): This is the basic_deliver.routing_key string
           in a rabbit message

        Returns:
           MessageCategory: the MessageCategory, as enum
        """
        routing_key_words = routing_key.split(".")
        msg_category_symbol_value = routing_key_words[0]

        try:
            msg_category_symbol = MessageCategorySymbol(msg_category_symbol_value)
        except ValueError:
            self._latest_on_message_diagnostic = (
                OnReceiveMessageDiagnostic.UNKNOWN_MESSAGE_CATEGORY_SYMBOL
            )
            raise GwTypeError(
                f"First  word of {routing_key} not a known MessageCategorySymbol!"
            )
        msg_category = utils.message_category_from_symbol(msg_category_symbol)
        allowable_categories = [
            MessageCategory.RabbitJsonDirect,
            MessageCategory.RabbitJsonBroadcast,
            MessageCategory.MqttJsonBroadcast,
        ]
        if msg_category not in allowable_categories:
            self._latest_on_message_diagnostic = (
                OnReceiveMessageDiagnostic.UNHANDLED_ROUTING_KEY_TYPE
            )
            raise GwTypeError(
                f"Rabbit messages do not handle {msg_category_symbol.value}"
            )
        return msg_category

    def type_name_from_routing_key(
        self,
        routing_key: str,
    ) -> str:
        """Returns the TypeName of the message given the routing key. Raises a GwTypeError
        exception if there is a problem decoding the TypeName, or if it does not have
        the appropriate left-right-dot format.

        Args:
            routing_key (str): This is the basic_deliver.routing_key string
            in a rabbit message

        Returns:
            str: the TypeName of the payload, in Lrd format
        """
        routing_key_words = routing_key.split(".")
        msg_category = self.message_category_from_routing_key(routing_key)
        if msg_category == MessageCategory.MqttJsonBroadcast:
            type_name_lrh = routing_key_words[2]
        else:
            type_name_lrh = routing_key_words[3]

        if not is_lrh_alias_format(type_name_lrh):
            self._latest_on_message_diagnostic = (
                OnReceiveMessageDiagnostic.TYPE_NAME_DECODING_PROBLEM
            )
            raise GwTypeError(
                f"TypeName {type_name_lrh} in {routing_key} message not lrh_alias_format!"
            )
        type_name = type_name_lrh.replace("-", ".")
        return type_name

    def from_alias_from_routing_key(self, routing_key: str) -> str:
        """Returns the GNodeAlias in left-right-dot format. Raises a GwTypeError
        if there is trouble getting this.
        Args:
            routing_key (str): This is the basic_deliver.routing_key string
            in a rabbit message
        """
        routing_key_words = routing_key.split(".")
        try:
            from_g_node_alias_lrh = routing_key_words[1]
        except:
            raise GwTypeError(f"{routing_key} must have at least two words!")

        if not is_lrh_alias_format(from_g_node_alias_lrh):
            raise GwTypeError(
                f"GNodeAlias not is_lrh_alias_format for routing key {routing_key}!"
            )
        from_g_node_alias = from_g_node_alias_lrh.replace("-", ".")
        return from_g_node_alias

    def from_role_from_routing_key(self, routing_key: str) -> GNodeRole:
        """Returns the GNodeRole that the message came from. Raises a GwTypeError
        if there is trouble getting this.
        Args:
            routing_key (str): basic_deliver.routing_key string
            in a rabbit message

        Raises:
            GwTypeError: Error in message construction

        Returns:
            GNodeRole: GNodeRole of the actor that sent the message
        """
        msg_category = self.message_category_from_routing_key(routing_key)
        if msg_category == MessageCategory.MqttJsonBroadcast:
            raise Exception(
                "MqttJsonBroadcast does not contain FromRole in routing key"
            )
        routing_key_words = routing_key.split(".")
        try:
            from_g_node_rabbit_role = routing_key_words[2]
        except:
            raise GwTypeError(f"{routing_key} must have at least three words!")
        try:
            rabbit_role = RabbitRole(from_g_node_rabbit_role)
        except ValueError:
            raise GwTypeError(
                f"Unknown short alias {from_g_node_rabbit_role} in {routing_key}"
                f" Must belong to {RabbitRole}"
            )

        return RoleByRabbitRole[rabbit_role]

    @no_type_check
    def send_message(
        self,
        payload: HeartbeatA,
        message_category: MessageCategory = MessageCategory.RabbitJsonDirect,
        to_role: Optional[GNodeRole] = None,
        to_g_node_alias: Optional[str] = None,
        radio_channel: Optional[str] = None,
    ) -> OnSendMessageDiagnostic:
        """Publish a direct message to another GNode in the registry world. The only type
        of direct messages in the registry use json (i.e. no more streamlined serial encoding),
        unlike in non-registry worlds.

        Args:
            payload: Any GridWorks types with a json content-type
            that includes TypeName as a json key, and has as_type()
            as an encoding method.
            routing_key_type: for creating routing key
            to_role (Optional[GNodeRole]): used if a direct message
            to_g_node_alias (str): used if a direct message

        Returns:
            OnSendMessageDiagnostic: MESSAGE_SENT with success, otherwise some
            description of why the message was not sent.
        """
        if self._stopping:
            return OnSendMessageDiagnostic.STOPPING_SO_NOT_SENDING
        if self._stopped:
            return OnSendMessageDiagnostic.STOPPED_SO_NOT_SENDING

        if "MessageId" in payload.as_dict():
            correlation
        else:
            correlation_id = str(uuid.uuid4())

        properties = pika.BasicProperties(
            reply_to=self.queue_name,
            app_id=self.alias,
            type=message_category,
            correlation_id=correlation_id,
        )
        publish_exchange = self._publish_exchange
        if message_category == MessageCategory.RabbitJsonDirect:
            if not isinstance(to_role, GNodeRole):
                raise Exception("Must include to_role for a direct message")
            try:
                check_is_left_right_dot(to_g_node_alias)
            except:
                raise Exception(
                    f"to_g_node_alias must have LrdAliasFormat. Got {to_g_node_alias}"
                )
            routing_key = self.direct_routing_key(
                to_role=to_role,
                payload=payload,
                to_g_node_alias=to_g_node_alias,
            )
        elif message_category == MessageCategory.RabbitJsonBroadcast:
            routing_key = self.broadcast_routing_key(
                payload=payload, radio_channel=radio_channel
            )
        elif message_category == MessageCategory.MqttJsonBroadcast:
            routing_key = self.scada_routing_key(
                payload=payload, to_g_node_alias=to_g_node_alias
            )
            publish_exchange = "amq.topic"
        else:
            raise Exception(f"Does not handle MessageCategory {message_category}")

        # if self._publish_channel is None:
        #     LOGGER.error(f"No publish channel so not sending {routing_key}")
        #     return OnSendMessageDiagnostic.CHANNEL_NOT_OPEN
        # if not self._publish_channel.is_open:
        #     LOGGER.error(f"Publish channel not open so not sending {routing_key}")
        #     return OnSendMessageDiagnostic.CHANNEL_NOT_OPEN

        if self._consume_channel is None:
            LOGGER.error(f"No channel so not sending {routing_key}")
            return OnSendMessageDiagnostic.CHANNEL_NOT_OPEN
        if not self._consume_channel.is_open:
            LOGGER.error(f"Channel not open so not sending {routing_key}")
            return OnSendMessageDiagnostic.CHANNEL_NOT_OPEN

        try:
            self._consume_channel.basic_publish(
                exchange=publish_exchange,
                routing_key=routing_key,
                body=payload.as_type(),
                properties=properties,
            )
            # self._publish_channel.basic_publish(
            #     exchange=self._publish_exchange,
            #     routing_key=routing_key,
            #     body=payload.as_type(),
            #     properties=properties,
            # )
            LOGGER.debug(f" [x] Sent {payload.type_name} w routing key {routing_key}")
            return OnSendMessageDiagnostic.MESSAGE_SENT

        except BaseException:
            LOGGER.exception("Problem publishing w consume channel")
            # LOGGER.exception("Problem w publish channel")
            return OnSendMessageDiagnostic.UNKNOWN_ERROR

    #################################################
    # Receiving messages
    #################################################

    @no_type_check
    def on_message(
        self,
        _unused_channel: PikaChannel,
        basic_deliver: Basic.Deliver,
        properties: BasicProperties,
        body: bytes,
    ) -> None:
        """
        Invoked by pika when a message is delivered from RabbitMQ. If a message
        does not get here that you expect should get here, check the routing key
        of the outbound message and the rabbitmq bindings.

        Parses the TypeName of the message payload and the GNodeAlias of the sender.
        If it recognizes the GNode and the TypeName, then it sends the message on to
        the check_routing function, which will be defined in a child class (e.g., the
        GNodeFactoryActorBase if the actor is a GNodeFactory).

        From RabbitMQ: The channel is passed for your convenience. The basic_deliver
        object that is passed in carries the exchange, delivery tag, and a redelivered
        flag for the message. The properties passed in is an instance of BasicProperties
        with the message properties including the routing key. The body is the message
        that was sent.

        :param pika.channel.Channel _unused_channel: The channel object
        :param pika.spec.Basic.Deliver basic_deliver: The basic.deliver method
        :param pika.spec.BasicProperties properties: The message properties including the routing key
        :param bytes body: The message body
        """
        self.latest_routing_key = basic_deliver.routing_key

        LOGGER.debug(
            f"{self.alias}: Got {basic_deliver.routing_key} with delivery tag {basic_deliver.delivery_tag}"
        )
        self.acknowledge_message(basic_deliver.delivery_tag)
        try:
            type_name = self.get_payload_type_name(basic_deliver)
        except GwTypeError:
            return

        if type_name not in self.api_type_maker_by_name.keys():
            self._latest_on_message_diagnostic = (
                OnReceiveMessageDiagnostic.UNKNOWN_TYPE_NAME
            )
            LOGGER.warning(
                f"IGNORING MESSAGE. {self._latest_on_message_diagnostic}: {type_name}"
            )
            return

        try:
            payload = self.api_type_maker_by_name[type_name].type_to_tuple(body)
        except Exception as e:
            LOGGER.warning(
                f"TypeName for incoming message claimed to be {type_name}, but was not true! Failed to make a {self.api_type_maker_by_name[type_name].tuple}"
            )
            return

        routing_key: str = basic_deliver.routing_key
        msg_category = self.message_category_from_routing_key(routing_key)
        try:
            from_alias = self.from_alias_from_routing_key(routing_key)
        except GwTypeError as e:
            self._latest_on_message_diagnostic = (
                OnReceiveMessageDiagnostic.FROM_GNODE_DECODING_PROBLEM
            )
            LOGGER.warning(
                f"IGNORING MESSAGE. {self._latest_on_message_diagnostic}: {e}"
            )
            return
        if msg_category == MessageCategory.MqttJsonBroadcast:
            self.route_mqtt_message(from_alias=from_alias, payload=payload)
            return
        try:
            from_role = self.from_role_from_routing_key(routing_key)
        except GwTypeError as e:
            self._latest_on_message_diagnostic = (
                OnReceiveMessageDiagnostic.FROM_GNODE_DECODING_PROBLEM
            )
            LOGGER.warning(
                f"IGNORING MESSAGE. {self._latest_on_message_diagnostic}: {e}"
            )
            return

        self._latest_on_message_diagnostic = (
            OnReceiveMessageDiagnostic.TO_DIRECT_ROUTING
        )
        self.route_message(
            from_alias=from_alias,
            from_role=from_role,
            payload=payload,
        )

    ########################
    ## Receives
    ########################

    def route_message(
        self, from_alias: str, from_role: GNodeRole, payload: SimTimestep
    ) -> None:
        """
        Base class for message routing in GNode Actors, handling interactions
        with the Supervisor and TimeCoordinator.

        Derived classes are expected to implement their own `route_message` method.
        It is recommended to call `super().route_message(from_alias, from_role, payload)`
        at the end of the method if the message has not been routed yet.
        """

        if payload.type_name == HeartbeatA_Maker.type_name:
            if from_role != GNodeRole.Supervisor:
                LOGGER.info(
                    f"Ignoring HeartbeatA from GNode {from_alias} with GNodeRole {from_role}; expects"
                    f"Supervisor as the GNodeRole"
                )
                return
            elif from_alias != self.settings.my_super_alias:
                LOGGER.info(
                    f"Ignoring HeartbeatA from supervisor {from_alias}; "
                    f"my supervisor is {self.settings.my_super_alias}"
                )
                return

            try:
                self.heartbeat_from_super(from_alias, payload)
            except:
                LOGGER.exception("Error in heartbeat_received")
        elif payload.type_name == SimTimestep_Maker.type_name:
            try:
                self.timestep_from_timecoordinator(payload)
            except:
                LOGGER.exception("Error in timestep_from_timecoordinator")

    def route_mqtt_message(self, from_alias: str, payload: HeartbeatA) -> None:
        """
        Base class for message routing from SCADA actors, which use
        MessageCategory.MqttJsonBroadcast

        This is only intended to be used for AtomicTNodes and Ears.
        """
        ...

    def heartbeat_from_super(self, from_alias: str, ping: HeartbeatA) -> None:
        """
        Subordinate GNode responds to its supervisor's heartbeat with a "pong" message.

        Both the received heartbeat (ping) and the response (pong) have the type HeartbeatA
        (see: https://gridworks.readthedocs.io/en/latest/apis/types.html#heartbeata).

        The subordinate GNode generates its own unique identifier (hex) and includes it
        in the pong message along with the heartbeat it received from the supervisor.

        Note that the subordinate GNode does not have the responsibility of verifying
        the authenticity of the last heartbeat received from the supervisor - although typically,
        the supervisor does send the last heartbeat from this GNode (except during the initial
        heartbeat exchange).

        Args:
            from_alias (str): the alias of the GNode that sent the ping.
            ping (HeartbeatA): the heartbeat sent.

        Raises:
        ValueError: If `from_alias` is not this GNode's Supervisor alias.

        """
        if from_alias != self.settings.my_super_alias:
            raise ValueError(
                f"from_alias {from_alias} does not match my supervisor"
                f" {self.settings.my_super_alias}. This message should"
                f"have been filtered out in the route_message method."
            )
        pong = HeartbeatA_Maker(
            my_hex=str(random.choice("0123456789abcdef")), your_last_hex=ping.MyHex
        ).tuple

        self.send_message(
            payload=pong,
            to_role=GNodeRole.Supervisor,
            to_g_node_alias=self.settings.my_super_alias,
        )

        LOGGER.debug(
            f"[{self.alias}] Sent HB: SuHex {pong.YourLastHex}, AtnHex {pong.MyHex}"
        )

    ########################
    ## Time related (simulated time)
    ########################

    def timestep_from_timecoordinator(self, payload: SimTimestep) -> None:
        if self._time < payload.TimeUnixS:
            self._time = payload.TimeUnixS
            self.new_timestep(payload)
            LOGGER.debug(f"Time is now {self.time_str()}")
        elif self._time == payload.TimeUnixS:
            self.repeat_timestep(payload)

    def new_timestep(self, payload: SimTimestep) -> None:
        LOGGER.info("New timestep")

    def repeat_timestep(self, payload: SimTimestep) -> None:
        LOGGER.info("Timestep received again in atn_actor_base")

    def time(self) -> float:
        if self.universe_type == UniverseType.Dev:
            return self._time
        else:
            return time.time()

    def time_str(self) -> str:
        timestamp = self.time()
        dt = datetime.datetime.fromtimestamp(timestamp, tz=datetime.timezone.utc)
        return dt.strftime("%m/%d/%Y, %H:%M")

    ###############################
    # Other GNode-related methods
    ###############################

    @property
    def short_alias(self) -> str:
        return self.alias.split(".")[-1]


def check_is_left_right_dot(v: str) -> None:
    """Checks LeftRightDot Format

    LeftRightDot format: Lowercase alphanumeric words separated by periods, with
    the most significant word (on the left) starting with an alphabet character.

    Args:
        v (str): the candidate

    Raises:
        ValueError: if v is not LeftRightDot format
    """
    from typing import List

    try:
        x: List[str] = v.split(".")
    except:
        raise ValueError(f"Failed to seperate <{v}> into words with split'.'")
    first_word = x[0]
    first_char = first_word[0]
    if not first_char.isalpha():
        raise ValueError(
            f"Most significant word of <{v}> must start with alphabet char."
        )
    for word in x:
        if not word.isalnum():
            raise ValueError(f"words of <{v}> split by by '.' must be alphanumeric.")
    if not v.islower():
        raise ValueError(f"All characters of <{v}> must be lowercase.")


def is_lrh_alias_format(candidate: str) -> bool:
    """AlphanumericStrings separated by hyphens, with most
    significant word to the left.  I.e. `d1.ne` is the child of `d1`.
    Checking the format cannot verify the significance of words. All
    words must be alphanumeric. Most significant word must start with
    an alphabet charecter"""
    try:
        x = candidate.split("-")
    except:
        return False
    first_word = x[0]
    first_char = first_word[0]
    if not first_char.isalpha():
        return False
    for word in x:
        if not word.isalnum():
            return False
    if not candidate.islower():
        return False
    return True



"""Settings for the GNodeFactory, readable from environment and/or from env files."""

import datetime

import algosdk
from pydantic import BaseModel
from pydantic import BaseSettings
from pydantic import SecretStr
from pydantic import root_validator
from pydantic import validator

from gwbase.enums import GNodeRole
from gwbase.enums import UniverseType


DEFAULT_ENV_FILE = ".env"


def check_is_algo_secret_key_format(v: str) -> None:
    try:
        b = algosdk.account.address_from_private_key(v)
    except Exception as e:
        raise ValueError(
            f"Not Algorand Secret Key format! Generate one by: \n"
            f"import algosdk\n"
            "private_key, address = algosdk.account.generate_account()\n"
            f"incorrect sk: {v}"
        )


def check_is_left_right_dot(v: str) -> None:
    """
    LeftRightDot format: Lowercase alphanumeric words separated by periods,
    most significant word (on the left) starting with an alphabet character.

    Raises:
        ValueError: if not LeftRightDot format
    """
    from typing import List

    try:
        x: List[str] = v.split(".")
    except:
        raise ValueError(f"Failed to seperate {v} into words with split'.'")
    first_word = x[0]
    first_char = first_word[0]
    if not first_char.isalpha():
        raise ValueError(f"Most significant word of {v} must start with alphabet char.")
    for word in x:
        if not word.isalnum():
            raise ValueError(f"words of {v} split by by '.' must be alphanumeric.")
    if not v.islower():
        raise ValueError(f"All characters of {v} must be lowercase.")


def check_is_reasonable_unix_time_s(v: int) -> None:
    """
    ReasonableUnixTimeS format: time in unix seconds between Jan 1 2000 and Jan 1 3000

    Raises:
        ValueError: if not ReasonableUnixTimeS format
    """
    from datetime import datetime
    from datetime import timezone

    start_date = datetime(2000, 1, 1, tzinfo=timezone.utc)
    end_date = datetime(3000, 1, 1, tzinfo=timezone.utc)

    start_timestamp = int(start_date.timestamp())
    end_timestamp = int(end_date.timestamp())

    if v < start_timestamp:
        raise ValueError(f"{v} must be after Jan 1 2000")
    if v > end_timestamp:
        raise ValueError(f"{v} must be before Jan 1 3000")


def check_g_node_alias(alias: str, universe_type: UniverseType) -> None:
    check_is_left_right_dot(alias)
    first_word = alias.split(".")[0]
    if universe_type == UniverseType.Dev:
        if not first_word.startswith("d"):
            raise ValueError(
                f"Bad alias {alias} ... Dev universe GNode alias must start with d."
            )
    elif universe_type == UniverseType.Hybrid:
        if not first_word.startswith("h"):
            raise ValueError(
                f"Bad alias {alias} ... Hybrid universe GNode alias must start with h."
            )
    elif universe_type == UniverseType.Production:
        if not first_word == "w":
            raise ValueError(
                f"Bad alias {alias} ... Production universe GNode alias must have w as first word."
            )


class Public(BaseModel):
    """
    Publicly available information about the GNodeFactory, including:
      - GnfAdminAddr
      - GnfApiRoot
      - TaValidatorFundingThresholdAlgos
      - TaDeedConsiderationAlgos

    Also includes useful information shartcuts for running the simulated Millinocket
    demo. In this demo there is only one MarketMaker, and one TaValidator.
    Public Algorand addresses and ApiRoots are included for both.
    """

    gnf_api_root: str = "http://localhost:8000"
    dev_market_maker_api_root: str = "http://localhost:7997"
    dev_ta_validator_api_root: str = "http://localhost:8001"
    gnf_admin_addr: str = "RNMHG32VTIHTC7W3LZOEPTDGREL5IQGK46HKD3KBLZHYQUCAKLMT4G5ALI"
    gnr_addr: str = "X2ASUAUPK5ICMGDXQZQKBPSXWEJLBA4KKQ2TXW2KWO2JQTLY3J2Q4S33WE"
    dev_market_maker_addr: str = (
        "JMEUH2AXM6UGRJO2DBZXDOA2OMIWQFNQZ54LCVC4GQX6QDOX5Z6JRGMWFA"
    )
    dev_ta_validator_addr: str = (
        "7QQT4GN3ZPAQEFCNWF5BMF7NULVK3CWICZVT4GM3BQRISD52YEDLWJ4MII"
    )
    dev_ta_validator_multi_addr: str = (
        "Y5TRQXIJHWJ4OHCZSWP4PZTCES5VWOF2KDTNYSMU5HLAUXBFQQDX6IR5KM'"
    )
    ta_validator_funding_threshold_algos: int = 100
    ta_deed_consideration_algos: int = 50
    universe: str = "dev"
    gnf_graveyard_addr: str = (
        "COA6SYUOBE33F5JDYEGC5XAD43QRG3VGHNNQXLYWFSSQEHDQ5HJ52NDNPI"
    )
    algod_address: str = "http://localhost:4001"
    kmd_address: str = "http://localhost:4002"
    gen_kmd_wallet_name: str = "unencrypted-default-wallet"


class AlgoApiSecrets(BaseModel):
    algod_token: SecretStr = SecretStr(
        "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"
    )
    kmd_token: SecretStr = SecretStr(
        "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"
    )
    gen_kmd_wallet_password: SecretStr = SecretStr("")


class VanillaSettings(BaseSettings):
    algo_api_secrets: AlgoApiSecrets = AlgoApiSecrets()
    public: Public = Public()

    class Config:
        env_prefix = "VANILLA_"
        env_nested_delimiter = "__"


class RabbitBrokerClient(BaseModel):
    """Settings for connecting to a Rabbit Broker"""

    url: SecretStr = SecretStr("amqp://smqPublic:smqPublic@localhost:5672/d1__1")


class GNodeSettings(BaseSettings):
    """
    Template settings for a GNode.
    """

    public: Public = Public()
    algo_api_secrets: AlgoApiSecrets = AlgoApiSecrets()
    rabbit: RabbitBrokerClient = RabbitBrokerClient()
    redis_endpoint: str = "localhost"

    g_node_alias: str = "d1.isone.unknown.gnode"
    g_node_id: str = "e23eb2ec-4064-4921-89d4-b006edc81216"
    g_node_instance_id: str = "00000000-0000-0000-0000-000000000000"
    g_node_role_value: str = "GNode"
    sk: SecretStr = SecretStr(
        "3g+IYDCVM84Ady7a8fGImRkEZ77+a4e3i14ub0QMjM/JKlzB2GNdv0S+lqMsYgPiGbd7aAp5943X5NzvdQJohw=="
    )
    # Secret key for public Algorand adress ZEVFZQOYMNO36RF6S2RSYYQD4IM3O63IBJ47PDOX4TOO65ICNCDUF4HMLE
    universe_type_value: str = "Dev"
    my_super_alias: str = "d1.super1"
    my_time_coordinator_alias: str = "d1.time"
    initial_time_unix_s: int = int(
        datetime.datetime(2020, 1, 1, 4, 20, tzinfo=datetime.timezone.utc).timestamp()
    )

    log_level: str = "INFO"
    minute_cron_file: str = "cron_last_minute.txt"
    hour_cron_file: str = "cron_last_hour.txt"
    day_cron_file: str = "cron_last_day.txt"

    @validator("initial_time_unix_s", pre=True)
    def convert_initial_time_unix_s(cls, value):
        return int(value)

    @root_validator
    def validate_setting_axioms(cls, values):
        """
        Validates the following:
            - universe_type_value belongs to the GridWorks UniverseType enum
            - g_node_role_value belongs to the GridWOrks GNodeRole enum
            - All GNodeAliases (self, my time coordinator, my super) have the correct
            LeftRightDot format and match the alias pattern for the universe type
            - sk.get_secret_value() has the format of an Algorand secret key
            - initial_time_unix_s is a reasonable unix time in ms
        """
        g_node_alias = values.get("g_node_alias")
        g_node_role_value = values.get("g_node_role_value")
        my_super_alias = values.get("my_super_alias")
        my_time_coordinator_alias = values.get("my_time_coordinator_alias")
        sk = values.get("sk")
        initial_time_unix_s = values.get("initial_time_unix_s")
        universe_type_value = values.get("universe_type_value")
        if universe_type_value not in UniverseType.values():
            raise ValueError("universe_type_value must belong to UniverseType.values()")
        universe_type = UniverseType(universe_type_value)
        if g_node_role_value not in GNodeRole.values():
            raise ValueError("g_node_role_value must belong to GNodeRole.values()")
        check_g_node_alias(alias=g_node_alias, universe_type=universe_type)
        check_g_node_alias(alias=my_time_coordinator_alias, universe_type=universe_type)
        check_g_node_alias(alias=my_super_alias, universe_type=universe_type)
        check_is_algo_secret_key_format(sk.get_secret_value())
        check_is_reasonable_unix_time_s(initial_time_unix_s)
        return values

    class Config:
        env_prefix = "GNODE_"
        env_nested_delimiter = "__"


class SupervisorSettings(BaseSettings):
    g_node_alias: str = "d1.isone.ver.keene.super1"
    g_node_id: str = "664a3250-ce51-4fe3-9ce9-a4b6416451fb"
    g_node_instance_id: str = "20e7edec-05e5-4152-bfec-ec21ddd2e3dd"
    supervisor_container_id: str = "995b0334-9940-424f-8fb1-4745e52ba295"
    g_node_role_value: str = "Supervisor"
    my_time_coordinator_alias = "d1.time"
    log_level: str = "INFO"
    universe_type_value: str = "Dev"
    world_instance_name: str = "d1__1"
    rabbit: RabbitBrokerClient = RabbitBrokerClient()

    class Config:
        env_prefix = "SUPER_"
        env_nested_delimiter = "__"



""" List of all the types used by the actor."""

from typing import Dict
from typing import List
from typing import no_type_check

from gwbase.types import BaseGNodeGt_Maker
from gwbase.types import GNodeGt_Maker
from gwbase.types import GNodeInstanceGt_Maker
from gwbase.types import HeartbeatA_Maker
from gwbase.types import Ready_Maker
from gwbase.types import SimTimestep_Maker
from gwbase.types import SuperStarter_Maker
from gwbase.types import SupervisorContainerGt_Maker


TypeMakerByName: Dict[str, HeartbeatA_Maker] = {}


@no_type_check
def type_makers() -> List[HeartbeatA_Maker]:
    return [
        BaseGNodeGt_Maker,
        GNodeGt_Maker,
        GNodeInstanceGt_Maker,
        HeartbeatA_Maker,
        Ready_Maker,
        SimTimestep_Maker,
        SuperStarter_Maker,
        SupervisorContainerGt_Maker,
    ]


for maker in type_makers():
    TypeMakerByName[maker.type_name] = maker


def version_by_type_name() -> Dict[str, str]:
    """
    Returns:
        Dict[str, str]: Keys are TypeNames, values are versions
    """

    v: Dict[str, str] = {
        "base.g.node.gt": "002",
        "g.node.gt": "002",
        "g.node.instance.gt": "000",
        "heartbeat.a": "100",
        "ready": "001",
        "sim.timestep": "000",
        "super.starter": "000",
        "supervisor.container.gt": "000",
    }

    return v


def status_by_versioned_type_name() -> Dict[str, str]:
    """
    Returns:
        Dict[str, str]: Keys are versioned TypeNames, values are type status
    """

    v: Dict[str, str] = {
        "base.g.node.gt.002": "Pending",
        "g.node.gt.002": "Pending",
        "g.node.instance.gt.000": "Pending",
        "heartbeat.a.100": "Pending",
        "ready.001": "Pending",
        "sim.timestep.000": "Pending",
        "super.starter.000": "Pending",
        "supervisor.container.gt.000": "Pending",
    }

    return v



from typing import Dict


class GpsPoint:
    by_id: Dict[str, "GpsPoint"] = {}

    def __new__(cls, gps_point_id: str, *args, **kwargs):  # type: ignore
        try:
            return cls.by_id[gps_point_id]
        except KeyError:
            instance = super().__new__(cls)
            cls.by_id[gps_point_id] = instance
            return instance

    def __init__(
        self,
        gps_point_id: str,
        lat: float,
        lon: float,
    ):
        self.gps_point_id = gps_point_id
        self.lat = lat
        self.lon = lon

    def __repr__(self) -> str:
        rs = f"GpsPoint lat: {self.lat}, lon: {self.lon} \r gps_point_id: {self.gps_point_id}"
        return rs
